from flask import Flask, request, jsonify
from connpy import configfile, node, nodes, hooks
from connpy.ai import ai as myai
from waitress import serve
import os
import signal

app = Flask(__name__)
conf = configfile()

PID_FILE1 = "/run/connpy.pid"
PID_FILE2 = "/tmp/connpy.pid"


@app.route("/")
def root():
    return jsonify({
        'message': 'Welcome to Connpy api',
        'version': '1.0',
        'documentation': 'https://fluzzi.github.io/connpy/'
    })

@app.route("/list_nodes", methods=["POST"])
def list_nodes():
    conf = app.custom_config
    case = conf.config["case"]
    try:
        data = request.get_json()
        filter = data["filter"]
        if not case:
            if isinstance(filter, list):
                filter = [item.lower() for item in filter]
            else:
                filter = filter.lower()
        output = conf._getallnodes(filter)
    except:
        output = conf._getallnodes()
    return jsonify(output)

@app.route("/get_nodes", methods=["POST"])
def get_nodes():
    conf = app.custom_config
    case = conf.config["case"]
    try:
        data = request.get_json()
        filter = data["filter"]
        if not case:
            if isinstance(filter, list):
                filter = [item.lower() for item in filter]
            else:
                filter = filter.lower()
        output = conf._getallnodesfull(filter)
    except:
        output = conf._getallnodesfull()
    return jsonify(output)

@app.route("/ask_ai", methods=["POST"])
def ask_ai():
    conf = app.custom_config
    data = request.get_json()
    input = data["input"]
    if "dryrun" in data:
        dryrun = data["dryrun"] 
    else:
        dryrun = False
    if "chat_history" in data:
        chat_history = data["chat_history"]
    else:
        chat_history = None
    ai = myai(conf)
    return ai.ask(input, dryrun, chat_history)

@app.route("/confirm", methods=["POST"])
def confirm():
    conf = app.custom_config
    data = request.get_json()
    input = data["input"]
    ai = myai(conf)
    return str(ai.confirm(input))

@app.route("/run_commands", methods=["POST"])
def run_commands():
    conf = app.custom_config
    data = request.get_json()
    case = conf.config["case"]
    mynodes = {}
    args = {}
    try:
        action = data["action"]
        nodelist = data["nodes"]
        args["commands"] = data["commands"]
        if action == "test":
            args["expected"] = data["expected"]
    except KeyError as e:
        error = "'{}' is mandatory".format(e.args[0])
        return({"DataError": error})
    if isinstance(nodelist, list):
        mynodes = conf.getitems(nodelist)
    else:
        if not case:
            nodelist = nodelist.lower()
        if nodelist.startswith("@"):
            mynodes = conf.getitem(nodelist)
        else:
            mynodes[nodelist] = conf.getitem(nodelist)

    mynodes = nodes(mynodes, config=conf)
    try:
        args["vars"] = data["vars"]
    except:
        pass
    try:
        options = data["options"]
        thisoptions = {k: v for k, v in options.items() if k in ["prompt", "parallel", "timeout"]}
        args.update(thisoptions)
    except:
        options = None
    if action == "run":
        output = mynodes.run(**args)
    elif action == "test":
        output = {}
        output["result"] = mynodes.test(**args)
        output["output"] = mynodes.output
    else:
        error = "Wrong action '{}'".format(action)
        return({"DataError": error})
    return output

@hooks.MethodHook
def stop_api():
    # Read the process ID (pid) from the file
    try:
        with open(PID_FILE1, "r") as f:
            pid = int(f.readline().strip())
            port = int(f.readline().strip())
        PID_FILE=PID_FILE1
    except:
        try:
            with open(PID_FILE2, "r") as f:
                pid = int(f.readline().strip())
                port = int(f.readline().strip())
            PID_FILE=PID_FILE2
        except:
            print("Connpy api server is not running.")
            return 
    # Send a SIGTERM signal to the process
    try:
        os.kill(pid, signal.SIGTERM)
    except:
        pass
    # Delete the PID file
    os.remove(PID_FILE)
    print(f"Server with process ID {pid} stopped.")
    return port

@hooks.MethodHook
def debug_api(port=8048):
    app.custom_config = configfile()
    app.run(debug=True, port=port)

@hooks.MethodHook
def start_server(port=8048):
    app.custom_config = configfile()
    serve(app, host='0.0.0.0', port=port)

@hooks.MethodHook
def start_api(port=8048):
    if os.path.exists(PID_FILE1) or os.path.exists(PID_FILE2):
        print("Connpy server is already running.")
        return
    pid = os.fork()
    if pid == 0:
        start_server(port)
    else:
        try:
            with open(PID_FILE1, "w") as f:
                f.write(str(pid) + "\n" + str(port))
        except:
            try:
                with open(PID_FILE2, "w") as f:
                    f.write(str(pid) + "\n" + str(port))
            except:
                print("Cound't create PID file")
                return
        print(f'Server is running with process ID {pid} in port {port}')




import openai
import time
import json
import re
import ast
from textwrap import dedent
from .core import nodes
from copy import deepcopy
from .hooks import ClassHook,MethodHook

@ClassHook
class ai:
    ''' This class generates a ai object. Containts all the information and methods to make requests to openAI chatGPT to run actions on the application.

    ### Attributes:  

        - model        (str): Model of GPT api to use. Default is gpt-3.5-turbo.

        - temp       (float): Value between 0 and 1 that control the randomness 
                              of generated text, with higher values increasing 
                              creativity. Default is 0.7.

        '''

    def __init__(self, config, org = None, api_key = None, model = None, temp = 0.7):
        ''' 
            
        ### Parameters:  

            - config (obj): Pass the object created with class configfile with 
                            key for decryption and extra configuration if you 
                            are using connection manager.  

        ### Optional Parameters:  

            - org     (str): A unique token identifying the user organization
                             to interact with the API.

            - api_key (str): A unique authentication token required to access 
                             and interact with the API.

            - model   (str): Model of GPT api to use. Default is gpt-3.5-turbo. 

            - temp  (float): Value between 0 and 1 that control the randomness 
                             of generated text, with higher values increasing 
                             creativity. Default is 0.7.
   

        '''
        self.config = config
        if org:
            openai.organization = org
        else:
            try: 
                openai.organization = self.config.config["openai"]["organization"]
            except:
                raise ValueError("Missing openai organization")
        if api_key:
            openai.api_key = api_key
        else:
            try: 
                openai.api_key = self.config.config["openai"]["api_key"]
            except:
                raise ValueError("Missing openai api_key")
        if model:
            self.model = model
        else:
            try:
                self.model = self.config.config["openai"]["model"]
            except:
                self.model = "gpt-3.5-turbo"
        self.temp = temp
        self.__prompt = {}
        self.__prompt["original_system"] = """
            You are the AI chatbot and assistant of a network connection manager and automation app called connpy. When provided with user input analyze the input and extract the following information. If user wants to chat just reply and don't call a function:

            - type: Given a user input, identify the type of request they want to make. The input will represent one of two options: 

                1. "command" - The user wants to get information from devices by running commands.
                2. "list_nodes" - The user wants to get a list of nodes, devices, servers, or routers.
                The 'type' field should reflect whether the user input is a command or a request for a list of nodes.

            - filter: One or more regex patterns indicating the device or group of devices the command should be run on. The filter can have different formats, such as:
                - hostname
                - hostname@folder
                - hostname@subfolder@folder
                - partofhostname
                - @folder
                - @subfolder@folder
                - regex_pattern

                The filter should be extracted from the user input exactly as it was provided.
                Always preserve the exact filter pattern provided by the user, with no modifications. Do not process any regex, the application can do that.

    """ 
        self.__prompt["original_user"] = "Get the IP addresses of loopback0 for all routers from w2az1 and e1.*(prod|dev) and check if they have the ip 192.168.1.1"
        self.__prompt["original_assistant"] = {"name": "get_network_device_info", "arguments": "{\n  \"type\": \"command\",\n  \"filter\": [\"w2az1\",\"e1.*(prod|dev)\"]\n}"}
        self.__prompt["original_function"] = {}
        self.__prompt["original_function"]["name"] = "get_network_device_info"
        self.__prompt["original_function"]["descriptions"] = "You are the AI chatbot and assistant of a network connection manager and automation app called connpy. When provided with user input analyze the input and extract the information acording to the function, If user wants to chat just reply and don't call a function",
        self.__prompt["original_function"]["parameters"] = {}
        self.__prompt["original_function"]["parameters"]["type"] = "object"
        self.__prompt["original_function"]["parameters"]["properties"] = {}
        self.__prompt["original_function"]["parameters"]["properties"]["type"] = {}
        self.__prompt["original_function"]["parameters"]["properties"]["type"]["type"] = "string"
        self.__prompt["original_function"]["parameters"]["properties"]["type"]["description"] ="""
Categorize the user's request based on the operation they want to perform on the nodes. The requests can be classified into the following categories:

    1. "command" - This represents a request to retrieve specific information or configurations from nodes. An example would be: "go to routers in @office and get the config".

    2. "list_nodes" - This is when the user wants a list of nodes. An example could be: "get me the nodes in @office".
"""
        self.__prompt["original_function"]["parameters"]["properties"]["type"]["enum"] = ["command", "list_nodes"]
        self.__prompt["original_function"]["parameters"]["properties"]["filter"] = {}
        self.__prompt["original_function"]["parameters"]["properties"]["filter"]["type"] = "array"
        self.__prompt["original_function"]["parameters"]["properties"]["filter"]["items"] = {}
        self.__prompt["original_function"]["parameters"]["properties"]["filter"]["items"]["type"] = "string"
        self.__prompt["original_function"]["parameters"]["properties"]["filter"]["items"]["description"] = """One or more regex patterns indicating the device or group of devices the command should be run on.  The filter should be extracted from the user input exactly as it was provided. 
                The filter can have different formats, such as:
                - hostname
                - hostname@folder
                - hostname@subfolder@folder
                - partofhostname
                - @folder
                - @subfolder@folder
                - regex_pattern
                """
        self.__prompt["original_function"]["parameters"]["required"] = ["type", "filter"]
        self.__prompt["command_system"] = """
        For each OS listed below, provide the command(s) needed to perform the specified action, depending on the device OS (e.g., Cisco IOSXR router, Linux server).
        The application knows how to connect to devices via SSH, so you only need to provide the command(s) to run after connecting. 
        If the commands needed are not for the specific OS type, just send an empty list (e.g., []). 
        Note: Preserving the integrity of user-provided commands is of utmost importance. If a user has provided a specific command to run, include that command exactly as it was given, even if it's not recognized or understood. Under no circumstances should you modify or alter user-provided commands.
    """
        self.__prompt["command_user"]= """
    input: show me the full configuration for all this devices:

    OS:
    cisco ios:
    """
        self.__prompt["command_assistant"] = {"name": "get_commands", "arguments": "{\n  \"cisco ios\": \"show running-configuration\"\n}"}
        self.__prompt["command_function"] = {}
        self.__prompt["command_function"]["name"] = "get_commands"
        self.__prompt["command_function"]["descriptions"] = """ 
        For each OS listed below, provide the command(s) needed to perform the specified action, depending on the device OS (e.g., Cisco IOSXR router, Linux server).
        The application knows how to connect to devices via SSH, so you only need to provide the command(s) to run after connecting. 
        If the commands needed are not for the specific OS type, just send an empty list (e.g., []). 
    """
        self.__prompt["command_function"]["parameters"] = {}
        self.__prompt["command_function"]["parameters"]["type"] = "object"
        self.__prompt["command_function"]["parameters"]["properties"] = {}
        self.__prompt["confirmation_system"] = """
        Please analyze the user's input and categorize it as either an affirmation or negation. Based on this analysis, respond with:

            'true' if the input is an affirmation like 'do it', 'go ahead', 'sure', etc.
            'false' if the input is a negation.
            'none' If the input does not fit into either of these categories.
            """
        self.__prompt["confirmation_user"] = "Yes go ahead!"
        self.__prompt["confirmation_assistant"] = "True"
        self.__prompt["confirmation_function"] = {}
        self.__prompt["confirmation_function"]["name"] = "get_confirmation"
        self.__prompt["confirmation_function"]["descriptions"] = """ 
        Analize user request and respond:
    """
        self.__prompt["confirmation_function"]["parameters"] = {}
        self.__prompt["confirmation_function"]["parameters"]["type"] = "object"
        self.__prompt["confirmation_function"]["parameters"]["properties"] = {}
        self.__prompt["confirmation_function"]["parameters"]["properties"]["result"] = {}
        self.__prompt["confirmation_function"]["parameters"]["properties"]["result"]["description"] = """'true' if the input is an affirmation like 'do it', 'go ahead', 'sure', etc.
'false' if the input is a negation.
'none' If the input does not fit into either of these categories"""
        self.__prompt["confirmation_function"]["parameters"]["properties"]["result"]["type"] = "string"
        self.__prompt["confirmation_function"]["parameters"]["properties"]["result"]["enum"] = ["true", "false", "none"]
        self.__prompt["confirmation_function"]["parameters"]["properties"]["response"] = {}
        self.__prompt["confirmation_function"]["parameters"]["properties"]["response"]["description"] = "If the user don't message is not an affiramtion or negation, kindly ask the user to rephrase."
        self.__prompt["confirmation_function"]["parameters"]["properties"]["response"]["type"] = "string"
        self.__prompt["confirmation_function"]["parameters"]["required"] = ["result"]

    @MethodHook
    def process_string(self, s):
        if s.startswith('[') and s.endswith(']') and not (s.startswith("['") and s.endswith("']")) and not (s.startswith('["') and s.endswith('"]')):
            # Extract the content inside square brackets and split by comma
            content = s[1:-1].split(',')
            # Add single quotes around each item and join them back together with commas
            new_content = ', '.join(f"'{item.strip()}'" for item in content)
            # Replace the old content with the new content
            s = '[' + new_content + ']'
        return s

    @MethodHook
    def _retry_function(self, function, max_retries, backoff_num, *args):
        #Retry openai requests
        retries = 0
        while retries < max_retries:
            try:
                myfunction = function(*args)
                break
            except:
                wait_time = backoff_num * (2 ** retries)
                time.sleep(wait_time)
                retries += 1
                continue
        if retries == max_retries:
            myfunction = False
        return myfunction

    @MethodHook
    def _clean_command_response(self, raw_response, node_list):
        #Parse response for command request to openAI GPT.
        info_dict = {}
        info_dict["commands"] = []
        info_dict["variables"] = {}
        info_dict["variables"]["__global__"] = {}
        for key, value in node_list.items():
            newvalue = {}
            commands = raw_response[value]
            for i,e in enumerate(commands, start=1):
                newvalue[f"command{i}"] = e
                if f"{{command{i}}}" not in info_dict["commands"]:
                    info_dict["commands"].append(f"{{command{i}}}")
                    info_dict["variables"]["__global__"][f"command{i}"] = ""
                info_dict["variables"][key] = newvalue
        return info_dict

    @MethodHook
    def _get_commands(self, user_input, nodes):
        #Send the request for commands for each device to openAI GPT.
        output_list = []
        command_function = deepcopy(self.__prompt["command_function"])
        node_list = {}
        for key, value in nodes.items():
            tags = value.get('tags', {})
            try:
                if os_value := tags.get('os'):
                    node_list[key] = os_value
                    output_list.append(f"{os_value}")
                    command_function["parameters"]["properties"][os_value] = {}
                    command_function["parameters"]["properties"][os_value]["type"] = "array"
                    command_function["parameters"]["properties"][os_value]["description"] = f"OS: {os_value}"
                    command_function["parameters"]["properties"][os_value]["items"] = {}
                    command_function["parameters"]["properties"][os_value]["items"]["type"] = "string" 
            except:
                pass
        output_str = "\n".join(list(set(output_list)))
        command_input = f"input: {user_input}\n\nOS:\n{output_str}"
        message = []
        message.append({"role": "system", "content": dedent(self.__prompt["command_system"]).strip()})
        message.append({"role": "user", "content": dedent(self.__prompt["command_user"]).strip()})
        message.append({"role": "assistant", "content": None, "function_call": self.__prompt["command_assistant"]})
        message.append({"role": "user", "content": command_input})
        functions = [command_function]
        response = openai.ChatCompletion.create(
            model=self.model,
            messages=message,
            functions=functions,
            function_call={"name": "get_commands"},
            temperature=self.temp
            )
        output = {}
        result = response["choices"][0]["message"].to_dict()
        json_result = json.loads(result["function_call"]["arguments"])
        output["response"] = self._clean_command_response(json_result, node_list)
        return output

    @MethodHook
    def _get_filter(self, user_input, chat_history = None):
        #Send the request to identify the filter and other attributes from the user input to GPT.
        message = []
        message.append({"role": "system", "content": dedent(self.__prompt["original_system"]).strip()})
        message.append({"role": "user", "content": dedent(self.__prompt["original_user"]).strip()})
        message.append({"role": "assistant", "content": None, "function_call": self.__prompt["original_assistant"]})
        functions = [self.__prompt["original_function"]]
        if not chat_history:
            chat_history = []
        chat_history.append({"role": "user", "content": user_input})
        message.extend(chat_history)
        response = openai.ChatCompletion.create(
            model=self.model,
            messages=message,
            functions=functions,
            function_call="auto",
            temperature=self.temp,
            top_p=1
            )

        def extract_quoted_strings(text):
            pattern = r'["\'](.*?)["\']'
            matches = re.findall(pattern, text)
            return matches
        expected = extract_quoted_strings(user_input)
        output = {}
        result = response["choices"][0]["message"].to_dict()
        if result["content"]:
            output["app_related"] = False
            chat_history.append({"role": "assistant", "content": result["content"]})
            output["response"] = result["content"]
        else:
            json_result = json.loads(result["function_call"]["arguments"])
            output["app_related"] = True
            output["filter"] = json_result["filter"]
            output["type"] = json_result["type"]
            chat_history.append({"role": "assistant", "content": result["content"], "function_call": {"name": result["function_call"]["name"], "arguments": json.dumps(json_result)}})
        output["expected"] = expected
        output["chat_history"] = chat_history
        return output
        
    @MethodHook
    def _get_confirmation(self, user_input):
        #Send the request to identify if user is confirming or denying the task
        message = []
        message.append({"role": "user", "content": user_input})
        functions = [self.__prompt["confirmation_function"]]
        response = openai.ChatCompletion.create(
            model=self.model,
            messages=message,
            functions=functions,
            function_call={"name": "get_confirmation"},
            temperature=self.temp,
            top_p=1
            )
        result = response["choices"][0]["message"].to_dict()
        json_result = json.loads(result["function_call"]["arguments"])
        output = {}
        if json_result["result"] == "true":
            output["result"] = True
        elif json_result["result"] == "false":
            output["result"] = False
        elif json_result["result"] == "none":
            output["result"] = json_result["response"]
        return output

    @MethodHook
    def confirm(self, user_input, max_retries=3, backoff_num=1):
        '''
        Send the user input to openAI GPT and verify if response is afirmative or negative.

        ### Parameters:  

            - user_input (str): User response confirming or denying.

        ### Optional Parameters:  

            - max_retries (int): Maximum number of retries for gpt api.
            - backoff_num (int): Backoff factor for exponential wait time
                                 between retries.

        ### Returns:  

            bool or str: True, False or str if AI coudn't understand the response
        '''
        result = self._retry_function(self._get_confirmation, max_retries, backoff_num, user_input)
        if result:
            output = result["result"]
        else:
            output = f"{self.model} api is not responding right now, please try again later."
        return output

    @MethodHook
    def ask(self, user_input, dryrun = False, chat_history = None,  max_retries=3, backoff_num=1):
        '''
        Send the user input to openAI GPT and parse the response to run an action in the application.

        ### Parameters:  

            - user_input (str): Request to send to openAI that will be parsed
                                and returned to execute on the application.
                                AI understands the following tasks:
                                - Run a command on a group of devices.
                                - List a group of devices.
                                - Test a command on a group of devices
                                  and verify if the output contain an
                                  expected value.

        ### Optional Parameters:  

            - dryrun       (bool): Set to true to get the arguments to use to
                                   run in the app. Default is false and it
                                   will run the actions directly.
            - chat_history (list): List in gpt api format for the chat history.
            - max_retries   (int): Maximum number of retries for gpt api.
            - backoff_num   (int): Backoff factor for exponential wait time
                                   between retries.

        ### Returns:  

            dict: Dictionary formed with the following keys:
                  - input: User input received
                  - app_related: True if GPT detected the request to be related
                    to the application.
                  - dryrun: True/False
                  - response: If the request is not related to the app. this
                    key will contain chatGPT answer.
                  - action: The action detected by the AI to run in the app.
                  - filter: If it was detected by the AI, the filter used
                    to get the list of nodes to work on.
                  - nodes: If it's not a dryrun, the list of nodes matched by
                    the filter.
                  - args: A dictionary of arguments required to run command(s)
                    on the nodes.
                  - result: A dictionary with the output of the commands or 
                    the test.
                  - chat_history: The chat history between user and chatbot.
                    It can be used as an attribute for next request.
                
                    

        '''
        output = {}
        output["dryrun"] = dryrun
        output["input"] = user_input
        original = self._retry_function(self._get_filter, max_retries, backoff_num, user_input, chat_history)
        if not original:
            output["app_related"] = False
            output["response"] = f"{self.model} api is not responding right now, please try again later."
            return output
        output["app_related"] = original["app_related"]
        output["chat_history"] = original["chat_history"]
        if not output["app_related"]:
            output["response"] = original["response"]
        else:
            type = original["type"]
            if "filter" in original:
                output["filter"] = original["filter"]
                if not self.config.config["case"]:
                    if isinstance(output["filter"], list):
                        output["filter"] = [item.lower() for item in output["filter"]]
                    else:
                        output["filter"] = output["filter"].lower()
                if not dryrun or type == "command":
                    thisnodes = self.config._getallnodesfull(output["filter"])
                    output["nodes"] = list(thisnodes.keys())
            if not type == "command":
                output["action"] = "list_nodes"
            else:
                if thisnodes:
                    commands = self._retry_function(self._get_commands, max_retries, backoff_num, user_input, thisnodes)
                else:
                    output["app_related"] = False
                    filterlist = ", ".join(output["filter"])
                    output["response"] = f"I'm sorry, I coudn't find any device with filter{'s' if len(output['filter']) != 1 else ''}: {filterlist}."
                    return output
                if not commands:
                    output["app_related"] = False
                    output["response"] = f"{self.model} api is not responding right now, please try again later."
                    return output
                output["args"] = {}
                output["args"]["commands"] = commands["response"]["commands"]
                output["args"]["vars"] = commands["response"]["variables"]
                output["nodes"] = [item for item in output["nodes"] if output["args"]["vars"].get(item)]
                if original.get("expected"):
                    output["args"]["expected"] = original["expected"]
                    output["action"] = "test"
                else:
                    output["action"] = "run"
                if dryrun:
                    output["task"] = []
                    if output["action"] == "test":
                        output["task"].append({"Task": "Verify if expected value is in command(s) output"})
                        output["task"].append({"Expected value to verify": output["args"]["expected"]})
                    elif output["action"] == "run":
                        output["task"].append({"Task": "Run command(s) on devices and return output"})
                    varstocommands = deepcopy(output["args"]["vars"])
                    del varstocommands["__global__"]
                    output["task"].append({"Devices": varstocommands})
                if not dryrun:
                    mynodes = nodes(self.config.getitems(output["nodes"]),config=self.config)
                    if output["action"] == "test":
                        output["result"] = mynodes.test(**output["args"])
                        output["logs"] = mynodes.output
                    elif output["action"] == "run":
                        output["result"] = mynodes.run(**output["args"])
        return output










#!/usr/bin/python3
import argparse
import os
import time
import zipfile
import tempfile
import io
import yaml
import threading
from google.oauth2.credentials import Credentials
from google.auth.transport.requests import Request
from googleapiclient.discovery import build
from google.auth.exceptions import RefreshError
from google_auth_oauthlib.flow import InstalledAppFlow
from googleapiclient.http import MediaFileUpload,MediaIoBaseDownload
from googleapiclient.errors import HttpError
from datetime import datetime

class sync:

    def __init__(self, connapp):
        self.scopes = ['https://www.googleapis.com/auth/drive.appdata']
        self.token_file = f"{connapp.config.defaultdir}/gtoken.json"
        self.file = connapp.config.file
        self.key = connapp.config.key
        self.google_client = f"{os.path.dirname(os.path.abspath(__file__))}/sync_client"
        self.connapp = connapp
        try:
            self.sync = self.connapp.config.config["sync"]
        except:
            self.sync = False

    def login(self):
        creds = None
        # The file token.json stores the user's access and refresh tokens.
        if os.path.exists(self.token_file):
            creds = Credentials.from_authorized_user_file(self.token_file, self.scopes)
        
        # If there are no valid credentials available, let the user log in.
        if not creds or not creds.valid:
            if creds and creds.expired and creds.refresh_token:
                creds.refresh(Request())
            else:
                flow = InstalledAppFlow.from_client_secrets_file(
                    self.google_client, self.scopes)
                creds = flow.run_local_server(port=0, access_type='offline')
            
            # Save the credentials for the next run
            with open(self.token_file, 'w') as token:
                token.write(creds.to_json())

        print("Logged in successfully.")

    def logout(self):
        if os.path.exists(self.token_file):
            os.remove(self.token_file)
            print("Logged out successfully.")
        else:
            print("No credentials file found. Already logged out.")

    def get_credentials(self):
        # Load credentials from token.json
        if os.path.exists(self.token_file):
            creds = Credentials.from_authorized_user_file(self.token_file, self.scopes)
        else:
            print("Credentials file not found.")
            return 0
        
        # If there are no valid credentials available, ask the user to log in again
        if not creds or not creds.valid:
            if creds and creds.expired and creds.refresh_token:
                try:
                    creds.refresh(Request())
                except RefreshError:
                    print("Could not refresh access token. Please log in again.")
                    return 0
            else:
                print("Credentials are missing or invalid. Please log in.")
                return 0
        return creds

    def check_login_status(self):
        # Check if the credentials file exists
        if os.path.exists(self.token_file):
            # Load credentials from token.json
            creds = Credentials.from_authorized_user_file(self.token_file)

            # If credentials are expired, refresh them
            if creds and creds.expired and creds.refresh_token:
                try:
                    creds.refresh(Request())
                except RefreshError:
                    pass

            # Check if the credentials are valid after refresh
            if creds.valid:
                return True
            else:
                return "Invalid"
        else:
            return False

    def status(self):
        print(f"Login: {self.check_login_status()}")
        print(f"Sync: {self.sync}")


    def get_appdata_files(self):

        creds = self.get_credentials()
        if not creds:
            return 0

        try:
            # Create the Google Drive service
            service = build("drive", "v3", credentials=creds)

            # List files in the appDataFolder
            response = (
                service.files()
                .list(
                    spaces="appDataFolder",
                    fields="files(id, name, appProperties)",
                    pageSize=10,
                )
                .execute()
            )

            files_info = []
            for file in response.get("files", []):
                # Extract file information
                file_id = file.get("id")
                file_name = file.get("name")
                timestamp = file.get("appProperties", {}).get("timestamp")
                human_readable_date = file.get("appProperties", {}).get("date")
                files_info.append({"name": file_name, "id": file_id, "date": human_readable_date, "timestamp": timestamp})

            return files_info

        except HttpError as error:
            print(f"An error occurred: {error}")
            return 0


    def dump_appdata_files_yaml(self):
        files_info = self.get_appdata_files()
        if not files_info:
            print("Failed to retrieve files or no files found.")
            return
        # Pretty print as YAML
        yaml_output = yaml.dump(files_info, sort_keys=False, default_flow_style=False)
        print(yaml_output)


    def backup_file_to_drive(self, file_path, timestamp):
        
        creds = self.get_credentials()
        if not creds:
            return 1
        
        # Create the Google Drive service
        service = build('drive', 'v3', credentials=creds)
        
        # Convert timestamp to a human-readable date
        human_readable_date = datetime.fromtimestamp(timestamp/1000).strftime('%Y-%m-%d %H:%M:%S')
        
        # Upload the file to Google Drive with timestamp metadata
        file_metadata = {
            'name': os.path.basename(file_path),
            'parents': ["appDataFolder"],
            'appProperties': {
                'timestamp': str(timestamp),
                'date': human_readable_date  # Add human-readable date attribute
            }
        }
        media = MediaFileUpload(file_path)
        
        try:
            file = service.files().create(body=file_metadata, media_body=media, fields='id').execute()
            return 0
        except Exception as e:
            return f"An error occurred: {e}"

    def delete_file_by_id(self, file_id):
        creds = self.get_credentials()
        if not creds:
            return 1

        try:
            # Create the Google Drive service
            service = build("drive", "v3", credentials=creds)

            # Delete the file
            service.files().delete(fileId=file_id).execute()
            return 0
        except Exception as e:
            return f"An error occurred: {e}"

    def compress_specific_files(self, zip_path):
        with zipfile.ZipFile(zip_path, 'w', zipfile.ZIP_DEFLATED) as zipf:
            zipf.write(self.file, "config.json")
            zipf.write(self.key, ".osk")

    def compress_and_upload(self):
        # Read the file content to get the folder path
        timestamp = int(time.time() * 1000)
        # Create a temporary directory for storing the zip file
        with tempfile.TemporaryDirectory() as tmp_dir:
            # Compress specific files from the folder path to a zip file in the temporary directory
            zip_path = os.path.join(tmp_dir, f"connpy-backup-{timestamp}.zip")
            self.compress_specific_files(zip_path)

            # Get the files in the app data folder
            app_data_files = self.get_appdata_files()
            if app_data_files == 0:
                return 1

            # If there are 10 or more files, remove the oldest one based on timestamp
            if len(app_data_files) >= 10:
                oldest_file = min(app_data_files, key=lambda x: x['timestamp'])
                delete_old = self.delete_file_by_id(oldest_file['id'])
                if delete_old:
                    print(delete_old)
                    return 1

            # Upload the new file
            upload_new = self.backup_file_to_drive(zip_path, timestamp)
            if upload_new:
                print(upload_new)
                return 1
            
            print("Backup to google uploaded successfully.")
            return 0

    def decompress_zip(self, zip_path):
        try:
            with zipfile.ZipFile(zip_path, 'r') as zipf:
                # Extract the specific file to the specified destination
                zipf.extract("config.json", os.path.dirname(self.file))
                zipf.extract(".osk", os.path.dirname(self.key))
            return 0
        except Exception as e:
            print(f"An error occurred: {e}")
            return 1

    def download_file_by_id(self, file_id, destination_path):

        creds = self.get_credentials()
        if not creds:
            return 1

        try:
            # Create the Google Drive service
            service = build('drive', 'v3', credentials=creds)

            # Download the file
            request = service.files().get_media(fileId=file_id)
            fh = io.FileIO(destination_path, mode='wb')
            downloader = MediaIoBaseDownload(fh, request)
            done = False
            while done is False:
                status, done = downloader.next_chunk()

            return 0
        except Exception as e:
            return f"An error occurred: {e}"

    def restore_last_config(self, file_id=None):
        # Get the files in the app data folder
        app_data_files = self.get_appdata_files()
        if not app_data_files:
            print("No files found in app data folder.")
            return 1

        # Check if a specific file_id was provided and if it exists in the list
        if file_id:
            selected_file = next((f for f in app_data_files if f['id'] == file_id), None)
            if not selected_file:
                print(f"No file found with ID: {file_id}")
                return 1
        else:
            # Find the latest file based on timestamp
            selected_file = max(app_data_files, key=lambda x: x['timestamp'])

        # Download the selected file to a temporary location
        temp_download_path = os.path.join(tempfile.gettempdir(), 'connpy-backup.zip')
        if self.download_file_by_id(selected_file['id'], temp_download_path):
            return 1

        # Unzip the downloaded file to the destination folder
        if self.decompress_zip(temp_download_path):
            print("Failed to decompress the file.")
            return 1

        print(f"Backup from Google Drive restored successfully: {selected_file['name']}")
        return 0

    def config_listener_post(self, args, kwargs):
        if self.sync:
            if self.check_login_status() == True:
                if not kwargs["result"]:
                    self.compress_and_upload()
        return kwargs["result"]

    def config_listener_pre(self, *args, **kwargs):
        try:
            self.sync = self.connapp.config.config["sync"]
        except:
            self.sync = False
        return args, kwargs

    def start_post_thread(self, *args, **kwargs):
        post_thread = threading.Thread(target=self.config_listener_post, args=(args,kwargs))
        post_thread.start()

class Preload:
    def __init__(self, connapp):
        syncapp = sync(connapp)
        connapp.config._saveconfig.register_post_hook(syncapp.start_post_thread)
        connapp.config._saveconfig.register_pre_hook(syncapp.config_listener_pre)

class Parser:
    def __init__(self):
        self.parser = argparse.ArgumentParser(description="Sync config with Google")
        self.description = "Sync config with Google"
        subparsers = self.parser.add_subparsers(title="Commands", dest='command',metavar="")
        login_parser = subparsers.add_parser("login", help="Login to Google to enable synchronization")
        logout_parser = subparsers.add_parser("logout", help="Logout from Google")
        start_parser = subparsers.add_parser("start", help="Start synchronizing with Google")
        stop_parser = subparsers.add_parser("stop", help="Stop any ongoing synchronization")
        restore_parser = subparsers.add_parser("restore", help="Restore data from Google")
        backup_parser = subparsers.add_parser("once", help="Backup current configuration to Google once")
        restore_parser.add_argument("--id", type=str, help="Optional file ID to restore a specific backup", required=False)
        status_parser = subparsers.add_parser("status", help="Check the current status of synchronization")
        list_parser = subparsers.add_parser("list", help="List all backups stored on Google")

class Entrypoint:
    def __init__(self, args, parser, connapp):
        syncapp = sync(connapp)
        if args.command == 'login':
            syncapp.login()
        elif args.command == "status":
            syncapp.status()
        elif args.command == "start":
            connapp._change_settings("sync", True)
        elif args.command == "stop":
            connapp._change_settings("sync", False)
        elif args.command == "list":
            syncapp.dump_appdata_files_yaml()
        elif args.command == "once":
            syncapp.compress_and_upload()
        elif args.command == "restore":
            syncapp.restore_last_config(args.id)
        elif args.command == "logout":
            syncapp.logout()

def _connpy_completion(wordsnumber, words, info = None):
    if wordsnumber == 3:
        result = ["--help", "login", "status", "start", "stop", "list", "once", "restore", "logout"]
    #NETMASK_completion
    if wordsnumber == 4 and words[1] == "restore":
            result = ["--help", "--id"]
    return result



#!/usr/bin/env python3
#Imports
import json
import os
import re
from Crypto.PublicKey import RSA
from pathlib import Path
from copy import deepcopy
from .hooks import MethodHook, ClassHook



#functions and classes

@ClassHook
class configfile:
    ''' This class generates a configfile object. Containts a dictionary storing, config, nodes and profiles, normaly used by connection manager.

    ### Attributes:  

        - file         (str): Path/file to config file.

        - key          (str): Path/file to RSA key file.

        - config      (dict): Dictionary containing information of connection
                              manager configuration.

        - connections (dict): Dictionary containing all the nodes added to
                              connection manager.

        - profiles    (dict): Dictionary containing all the profiles added to
                              connection manager.

        - privatekey   (obj): Object containing the private key to encrypt 
                              passwords.

        - publickey    (obj): Object containing the public key to decrypt 
                              passwords.
        '''

    def __init__(self, conf = None, key = None):
        ''' 
            
        ### Optional Parameters:  

            - conf (str): Path/file to config file. If left empty default
                          path is ~/.config/conn/config.json

            - key  (str): Path/file to RSA key file. If left empty default
                          path is ~/.config/conn/.osk

        '''
        home = os.path.expanduser("~")
        defaultdir = home + '/.config/conn'
        self.defaultdir = defaultdir
        Path(defaultdir).mkdir(parents=True, exist_ok=True)
        Path(f"{defaultdir}/plugins").mkdir(parents=True, exist_ok=True)
        pathfile = defaultdir + '/.folder'
        try:
            with open(pathfile, "r") as f:
                configdir = f.read().strip()
        except:
            with open(pathfile, "w") as f:
                f.write(str(defaultdir))
            configdir = defaultdir
        defaultfile = configdir + '/config.json'
        defaultkey = configdir + '/.osk'
        if conf == None:
            self.file = defaultfile
        else:
            self.file = conf
        if key == None:
            self.key = defaultkey
        else:
            self.key = key
        if os.path.exists(self.file):
            config = self._loadconfig(self.file)
        else:
            config = self._createconfig(self.file)
        self.config = config["config"]
        self.connections = config["connections"]
        self.profiles = config["profiles"]
        if not os.path.exists(self.key):
            self._createkey(self.key)
        with open(self.key) as f:
            self.privatekey = RSA.import_key(f.read())
            f.close()
        self.publickey = self.privatekey.publickey()


    def _loadconfig(self, conf):
        #Loads config file
        jsonconf = open(conf)
        jsondata = json.load(jsonconf)
        jsonconf.close()
        return jsondata

    def _createconfig(self, conf):
        #Create config file
        defaultconfig = {'config': {'case': False, 'idletime': 30, 'fzf': False}, 'connections': {}, 'profiles': { "default": { "host":"", "protocol":"ssh", "port":"", "user":"", "password":"", "options":"", "logs":"", "tags": "", "jumphost":""}}}
        if not os.path.exists(conf):
            with open(conf, "w") as f:
                json.dump(defaultconfig, f, indent = 4)
                f.close()
                os.chmod(conf, 0o600)
        jsonconf = open(conf)
        jsondata = json.load(jsonconf)
        jsonconf.close()
        return jsondata

    @MethodHook
    def _saveconfig(self, conf):
        #Save config file
        newconfig = {"config":{}, "connections": {}, "profiles": {}}
        newconfig["config"] = self.config
        newconfig["connections"] = self.connections
        newconfig["profiles"] = self.profiles
        try:
            with open(conf, "w") as f:
                json.dump(newconfig, f, indent = 4)
                f.close()
        except:
            return 1
        return 0

    def _createkey(self, keyfile):
        #Create key file
        key = RSA.generate(2048)
        with open(keyfile,'wb') as f:
            f.write(key.export_key('PEM'))
            f.close()
            os.chmod(keyfile, 0o600)
        return key

    @MethodHook
    def _explode_unique(self, unique):
        #Divide unique name into folder, subfolder and id
        uniques = unique.split("@")
        if not unique.startswith("@"):
            result = {"id": uniques[0]}
        else:
            result = {}
        if len(uniques) == 2:
            result["folder"] = uniques[1]
            if result["folder"] == "":
                return False
        elif len(uniques) == 3:
            result["folder"] = uniques[2]
            result["subfolder"] = uniques[1]
            if result["folder"] == "" or result["subfolder"] == "":
                return False
        elif len(uniques) > 3:
            return False
        return result

    @MethodHook
    def getitem(self, unique, keys = None):
        '''
        Get an node or a group of nodes from configfile which can be passed to node/nodes class

        ### Parameters:  

            - unique (str): Unique name of the node or folder in config using
                            connection manager style: node[@subfolder][@folder]
                            or [@subfolder]@folder

        ### Optional Parameters:  

            - keys (list): In case you pass a folder as unique, you can filter
                           nodes inside the folder passing a list.

        ### Returns:  

            dict: Dictionary containing information of node or multiple 
                  dictionaries of multiple nodes.

        '''
        uniques = self._explode_unique(unique)
        if unique.startswith("@"):
            if uniques.keys() >= {"folder", "subfolder"}:
                folder = self.connections[uniques["folder"]][uniques["subfolder"]]
            else:
                folder = self.connections[uniques["folder"]]
            newfolder = deepcopy(folder)
            newfolder.pop("type")
            for node in folder.keys():
                if node == "type":
                    continue
                if "type" in newfolder[node].keys():
                    if newfolder[node]["type"] == "subfolder":
                        newfolder.pop(node)
                    else:
                        newfolder[node].pop("type")
            if keys == None:
                newfolder = {"{}{}".format(k,unique):v for k,v in newfolder.items()}
                return newfolder
            else:
                f_newfolder = dict((k, newfolder[k]) for k in keys)
                f_newfolder = {"{}{}".format(k,unique):v for k,v in f_newfolder.items()}
                return f_newfolder
        else:
            if uniques.keys() >= {"folder", "subfolder"}:
                node = self.connections[uniques["folder"]][uniques["subfolder"]][uniques["id"]]
            elif "folder" in uniques.keys():
                node = self.connections[uniques["folder"]][uniques["id"]]
            else:
                node = self.connections[uniques["id"]]
            newnode = deepcopy(node)
            newnode.pop("type")
            return newnode

    @MethodHook
    def getitems(self, uniques):
        '''
        Get a group of nodes from configfile which can be passed to node/nodes class

        ### Parameters:  

            - uniques (str/list): String name that will match hostnames 
                                  from the connection manager. It can be a 
                                  list of strings.

        ### Returns:  

            dict: Dictionary containing information of node or multiple 
                  dictionaries of multiple nodes.

        '''
        nodes = {}
        if isinstance(uniques, str):
            uniques = [uniques]
        for i in uniques:
            if isinstance(i, dict):
                name = list(i.keys())[0]
                mylist = i[name]
                if not self.config["case"]:
                    name = name.lower()
                    mylist = [item.lower() for item in mylist]
                this = self.getitem(name, mylist)
                nodes.update(this)
            elif i.startswith("@"):
                if not self.config["case"]:
                    i = i.lower()
                this = self.getitem(i)
                nodes.update(this)
            else:
                if not self.config["case"]:
                    i = i.lower()
                this = self.getitem(i)
                nodes[i] = this
        return nodes


    @MethodHook
    def _connections_add(self,*, id, host, folder='', subfolder='', options='', logs='', password='', port='', protocol='', user='', tags='', jumphost='', type = "connection" ):
        #Add connection from config
        if folder == '':
            self.connections[id] = {"host": host, "options": options, "logs": logs, "password": password, "port": port, "protocol": protocol, "user": user, "tags": tags,"jumphost": jumphost,"type": type}
        elif folder != '' and subfolder == '':
            self.connections[folder][id] = {"host": host, "options": options, "logs": logs, "password": password, "port": port, "protocol": protocol, "user": user, "tags": tags, "jumphost": jumphost, "type": type}
        elif folder != '' and subfolder != '':
            self.connections[folder][subfolder][id] = {"host": host, "options": options, "logs": logs, "password": password, "port": port, "protocol": protocol, "user": user, "tags": tags,  "jumphost": jumphost, "type": type}
            

    @MethodHook
    def _connections_del(self,*, id, folder='', subfolder=''):
        #Delete connection from config
        if folder == '':
            del self.connections[id]
        elif folder != '' and subfolder == '':
            del self.connections[folder][id]
        elif folder != '' and subfolder != '':
            del self.connections[folder][subfolder][id]

    @MethodHook
    def _folder_add(self,*, folder, subfolder = ''):
        #Add Folder from config
        if subfolder == '':
            if folder not in self.connections:
                self.connections[folder] = {"type": "folder"}
        else:
            if subfolder not in self.connections[folder]:
                self.connections[folder][subfolder] = {"type": "subfolder"}

    @MethodHook
    def _folder_del(self,*, folder, subfolder=''):
        #Delete folder from config
        if subfolder == '':
            del self.connections[folder]
        else:
            del self.connections[folder][subfolder]


    @MethodHook
    def _profiles_add(self,*, id, host = '', options='', logs='', password='', port='', protocol='', user='', tags='', jumphost='' ):
        #Add profile from config
        self.profiles[id] = {"host": host, "options": options, "logs": logs, "password": password, "port": port, "protocol": protocol, "user": user, "tags": tags, "jumphost": jumphost}
            

    @MethodHook
    def _profiles_del(self,*, id ):
        #Delete profile from config
        del self.profiles[id]
        
    @MethodHook
    def _getallnodes(self, filter = None):
        #get all nodes on configfile
        nodes = []
        layer1 = [k for k,v in self.connections.items() if isinstance(v, dict) and v["type"] == "connection"]
        folders = [k for k,v in self.connections.items() if isinstance(v, dict) and v["type"] == "folder"]
        nodes.extend(layer1)
        for f in folders:
            layer2 = [k + "@" + f for k,v in self.connections[f].items() if isinstance(v, dict) and v["type"] == "connection"]
            nodes.extend(layer2)
            subfolders = [k for k,v in self.connections[f].items() if isinstance(v, dict) and v["type"] == "subfolder"]
            for s in subfolders:
                layer3 = [k + "@" + s + "@" + f for k,v in self.connections[f][s].items() if isinstance(v, dict) and v["type"] == "connection"]
                nodes.extend(layer3)
        if filter:
            if isinstance(filter, str):
                nodes = [item for item in nodes if re.search(filter, item)]
            elif isinstance(filter, list):
                nodes = [item for item in nodes if any(re.search(pattern, item) for pattern in filter)]
            else:
                raise ValueError("filter must be a string or a list of strings")
        return nodes

    @MethodHook
    def _getallnodesfull(self, filter = None, extract = True):
        #get all nodes on configfile with all their attributes.
        nodes = {}
        layer1 = {k:v for k,v in self.connections.items() if isinstance(v, dict) and v["type"] == "connection"}
        folders = [k for k,v in self.connections.items() if isinstance(v, dict) and v["type"] == "folder"]
        nodes.update(layer1)
        for f in folders:
            layer2 = {k + "@" + f:v for k,v in self.connections[f].items() if isinstance(v, dict) and v["type"] == "connection"}
            nodes.update(layer2)
            subfolders = [k for k,v in self.connections[f].items() if isinstance(v, dict) and v["type"] == "subfolder"]
            for s in subfolders:
                layer3 = {k + "@" + s + "@" + f:v for k,v in self.connections[f][s].items() if isinstance(v, dict) and v["type"] == "connection"}
                nodes.update(layer3)
        if filter:
            if isinstance(filter, str):
                filter = "^(?!.*@).+$" if filter == "@" else filter
                nodes = {k: v for k, v in nodes.items() if re.search(filter, k)}
            elif isinstance(filter, list):
                filter = ["^(?!.*@).+$" if item == "@" else item for item in filter]
                nodes = {k: v for k, v in nodes.items() if any(re.search(pattern, k) for pattern in filter)}
            else:
                raise ValueError("filter must be a string or a list of strings")
        if extract:
            for node, keys in nodes.items():
                for key, value in keys.items():
                    profile = re.search("^@(.*)", str(value))
                    if profile:
                        try:
                            nodes[node][key] = self.profiles[profile.group(1)][key]
                        except:
                            nodes[node][key] = ""
                    elif value == '' and key == "protocol":
                        try:
                            nodes[node][key] = config.profiles["default"][key]
                        except:
                            nodes[node][key] = "ssh"
        return nodes


    @MethodHook
    def _getallfolders(self):
        #get all folders on configfile
        folders = ["@" + k for k,v in self.connections.items() if isinstance(v, dict) and v["type"] == "folder"]
        subfolders = []
        for f in folders:
            s = ["@" + k + f for k,v in self.connections[f[1:]].items() if isinstance(v, dict) and v["type"] == "subfolder"]
            subfolders.extend(s)
        folders.extend(subfolders)
        return folders

    @MethodHook
    def _profileused(self, profile):
        #Check if profile is used before deleting it
        nodes = []
        layer1 = [k for k,v in self.connections.items() if isinstance(v, dict) and v["type"] == "connection" and ("@" + profile in v.values() or ( isinstance(v["password"],list) and "@" + profile in v["password"]))]
        folders = [k for k,v in self.connections.items() if isinstance(v, dict) and v["type"] == "folder"]
        nodes.extend(layer1)
        for f in folders:
            layer2 = [k + "@" + f for k,v in self.connections[f].items() if isinstance(v, dict) and v["type"] == "connection" and ("@" + profile in v.values() or ( isinstance(v["password"],list) and "@" + profile in v["password"]))]
            nodes.extend(layer2)
            subfolders = [k for k,v in self.connections[f].items() if isinstance(v, dict) and v["type"] == "subfolder"]
            for s in subfolders:
                layer3 = [k + "@" + s + "@" + f for k,v in self.connections[f][s].items() if isinstance(v, dict) and v["type"] == "connection" and ("@" + profile in v.values() or ( isinstance(v["password"],list) and "@" + profile in v["password"]))]
                nodes.extend(layer3)
        return nodes




#!/usr/bin/env python3
'''
## Connection manager

Connpy is a connection manager that allows you to store nodes to connect them fast and password free.

### Features
    - You can generate profiles and reference them from nodes using @profilename so you dont
      need to edit multiple nodes when changing password or other information.
    - Nodes can be stored on @folder or @subfolder@folder to organize your devices. Then can 
      be referenced using node@subfolder@folder or node@folder
    - If you have too many nodes. Get completion script using: conn config --completion.
      Or use fzf installing pyfzf and running conn config --fzf true
    - Create in bulk, copy, move, export and import nodes for easy management.
    - Run automation scripts in network devices.
    - use GPT AI to help you manage your devices.
    - Add plugins with your own scripts.
    - Much more!

### Usage
```
usage: conn [-h] [--add | --del | --mod | --show | --debug] [node|folder] [--sftp]
       conn {profile,move,mv,copy,cp,list,ls,bulk,export,import,ai,run,api,plugin,config} ...

positional arguments:
  node|folder    node[@subfolder][@folder]
                 Connect to specific node or show all matching nodes
                 [@subfolder][@folder]
                 Show all available connections globaly or in specified path
Options:
  -h, --help         show this help message and exit
  -v, --version      Show version
  -a, --add          Add new node[@subfolder][@folder] or [@subfolder]@folder
  -r, --del, --rm    Delete node[@subfolder][@folder] or [@subfolder]@folder
  -e, --mod, --edit  Modify node[@subfolder][@folder]
  -s, --show         Show node[@subfolder][@folder]
  -d, --debug        Display all conections steps
  -t, --sftp         Connects using sftp instead of ssh

Commands:
  profile         Manage profiles
  move(mv)        Move node
  copy(cp)        Copy node
  list(ls)        List profiles, nodes or folders
  bulk            Add nodes in bulk
  export          Export connection folder to Yaml file
  import          Import connection folder to config from Yaml file
  ai              Make request to an AI
  run             Run scripts or commands on nodes
  api             Start and stop connpy api
  plugin          Manage plugins
  config          Manage app config
  sync            Sync config with Google
```

###   Manage profiles
```
usage: conn profile [-h] (--add | --del | --mod | --show) profile

positional arguments:
  profile        Name of profile to manage

options:
  -h, --help         show this help message and exit
  -a, --add          Add new profile
  -r, --del, --rm    Delete profile
  -e, --mod, --edit  Modify profile
  -s, --show         Show profile

```

###   Examples
```
   conn profile --add office-user
   conn --add @office
   conn --add @datacenter@office
   conn --add server@datacenter@office
   conn --add pc@office
   conn --show server@datacenter@office
   conn pc@office
   conn server
``` 
## Plugin Requirements for Connpy
### General Structure
- The plugin script must be a Python file.
- Only the following top-level elements are allowed in the plugin script:
  - Class definitions
  - Function definitions
  - Import statements
  - The `if __name__ == "__main__":` block for standalone execution
  - Pass statements

### Specific Class Requirements
- The plugin script must define specific classes with particular attributes and methods. Each class serves a distinct role within the plugin's architecture:
  1. **Class `Parser`**:
     - **Purpose**: Handles parsing of command-line arguments.
     - **Requirements**:
       - Must contain only one method: `__init__`.
       - The `__init__` method must initialize at least two attributes:
         - `self.parser`: An instance of `argparse.ArgumentParser`.
         - `self.description`: A string containing the description of the parser.
  2. **Class `Entrypoint`**:
     - **Purpose**: Acts as the entry point for plugin execution, utilizing parsed arguments and integrating with the main application.
     - **Requirements**:
       - Must have an `__init__` method that accepts exactly three parameters besides `self`:
         - `args`: Arguments passed to the plugin.
         - The parser instance (typically `self.parser` from the `Parser` class).
         - The Connapp instance to interact with the Connpy app.
  3. **Class `Preload`**:
     - **Purpose**: Performs any necessary preliminary setup or configuration independent of the main parsing and entry logic.
   - **Requirements**:
     - Contains at least an `__init__` method that accepts parameter connapp besides `self`.

### Class Dependencies and Combinations
- **Dependencies**:
  - `Parser` and `Entrypoint` are interdependent and must both be present if one is included.
  - `Preload` is independent and may exist alone or alongside the other classes.
- **Valid Combinations**:
  - `Parser` and `Entrypoint` together.
  - `Preload` alone.
  - All three classes (`Parser`, `Entrypoint`, `Preload`).

### Preload Modifications and Hooks

In the `Preload` class of the plugin system, you have the ability to customize the behavior of existing classes and methods within the application through a robust hooking system. This documentation explains how to use the `modify`, `register_pre_hook`, and `register_post_hook` methods to tailor plugin functionality to your needs.

#### Modifying Classes with `modify`
The `modify` method allows you to alter instances of a class at the time they are created or after their creation. This is particularly useful for setting or modifying configuration settings, altering default behaviors, or adding new functionalities to existing classes without changing the original class definitions.

- **Usage**: Modify a class to include additional configurations or changes
- **Modify Method Signature**:
  - `modify(modification_method)`: A function that is invoked with an instance of the class as its argument. This function should perform any modifications directly on this instance.
- **Modification Method Signature**:
  - **Arguments**:
    - `cls`:  This function accepts a single argument, the class instance, which it then modifies.
  - **Modifiable Classes**:
    - `connapp.config`
    - `connapp.node`
    - `connapp.nodes`
    - `connapp.ai`
  - ```python
    def modify_config(cls):
        # Example modification: adding a new attribute or modifying an existing one
        cls.new_attribute = 'New Value'

    class Preload:
        def __init__(self, connapp):
            # Applying modification to the config class instance
            connapp.config.modify(modify_config)
    ```

#### Implementing Hooks with `register_pre_hook` and `register_post_hook`
These methods allow you to define custom logic to be executed before (`register_pre_hook`) or after (`register_post_hook`) the main logic of a method. This is particularly useful for logging, auditing, preprocessing inputs, postprocessing outputs or adding functionalities.

  - **Usage**: Register hooks to methods to execute additional logic before or after the main method execution.
- **Registration Methods Signature**:
  - `register_pre_hook(pre_hook_method)`: A function that is invoked before the main method is executed. This function should do preprocessing of the arguments.
  - `register_post_hook(post_hook_method)`: A function that is invoked after the main method is executed. This function should do postprocessing of the outputs.
- **Method Signatures for Pre-Hooks**
  - `pre_hook_method(*args, **kwargs)`
  - **Arguments**:
    - `*args`, `**kwargs`: The arguments and keyword arguments that will be passed to the method being hooked. The pre-hook function has the opportunity to inspect and modify these arguments before they are passed to the main method.
  - **Return**:
    - Must return a tuple `(args, kwargs)`, which will be used as the new arguments for the main method. If the original arguments are not modified, the function should return them as received.
- **Method Signatures for Post-Hooks**:
  - `post_hook_method(*args, **kwargs)`
  - **Arguments**:
    - `*args`, `**kwargs`: The arguments and keyword arguments that were passed to the main method.
      - `kwargs["result"]`: The value returned by the main method. This allows the post-hook to inspect and even alter the result before it is returned to the original caller.
  - **Return**:
    - Can return a modified result, which will replace the original result of the main method, or simply return `kwargs["result"]` to return the original method result.    
  - ```python
    def pre_processing_hook(*args, **kwargs):
        print("Pre-processing logic here")
        # Modify arguments or perform any checks
        return args, kwargs  # Return modified or unmodified args and kwargs

    def post_processing_hook(*args, **kwargs):
        print("Post-processing logic here")
        # Modify the result or perform any final logging or cleanup
        return kwargs["result"]  # Return the modified or unmodified result

    class Preload:
        def __init__(self, connapp):
            # Registering a pre-hook
            connapp.ai.some_method.register_pre_hook(pre_processing_hook)

            # Registering a post-hook
            connapp.node.another_method.register_post_hook(post_processing_hook)
    ```

### Executable Block
- The plugin script can include an executable block:
  - `if __name__ == "__main__":`
  - This block allows the plugin to be run as a standalone script for testing or independent use.

### Script Verification
- The `verify_script` method in `plugins.py` is used to check the plugin script's compliance with these standards.
- Non-compliant scripts will be rejected to ensure consistency and proper functionality within the plugin system.
- 
### Example Script

For a practical example of how to write a compatible plugin script, please refer to the following example:

[Example Plugin Script](https://github.com/fluzzi/awspy)

This script demonstrates the required structure and implementation details according to the plugin system's standards.

## http API
With the Connpy API you can run commands on devices using http requests

### 1. List Nodes

**Endpoint**: `/list_nodes`

**Method**: `POST`

**Description**: This route returns a list of nodes. It can also filter the list based on a given keyword.

#### Request Body:

```json
{
  "filter": "<keyword>"
}
```

* `filter` (optional): A keyword to filter the list of nodes. It returns only the nodes that contain the keyword. If not provided, the route will return the entire list of nodes.

#### Response:

- A JSON array containing the filtered list of nodes.

---

### 2. Get Nodes

**Endpoint**: `/get_nodes`

**Method**: `POST`

**Description**: This route returns a dictionary of nodes with all their attributes. It can also filter the nodes based on a given keyword.

#### Request Body:

```json
{
  "filter": "<keyword>"
}
```

* `filter` (optional): A keyword to filter the nodes. It returns only the nodes that contain the keyword. If not provided, the route will return the entire list of nodes.

#### Response:

- A JSON array containing the filtered nodes.

---

### 3. Run Commands

**Endpoint**: `/run_commands`

**Method**: `POST`

**Description**: This route runs commands on selected nodes based on the provided action, nodes, and commands. It also supports executing tests by providing expected results.

#### Request Body:

```json
{
  "action": "<action>",
  "nodes": "<nodes>",
  "commands": "<commands>",
  "expected": "<expected>",
  "options": "<options>"
}
```

* `action` (required): The action to be performed. Possible values: `run` or `test`.
* `nodes` (required): A list of nodes or a single node on which the commands will be executed. The nodes can be specified as individual node names or a node group with the `@` prefix. Node groups can also be specified as arrays with a list of nodes inside the group.
* `commands` (required): A list of commands to be executed on the specified nodes.
* `expected` (optional, only used when the action is `test`): A single expected result for the test.
* `options` (optional): Array to pass options to the run command, options are: `prompt`, `parallel`, `timeout`  

#### Response:

- A JSON object with the results of the executed commands on the nodes.

---

### 4. Ask AI

**Endpoint**: `/ask_ai`

**Method**: `POST`

**Description**: This route sends to chatgpt IA a request that will parse it into an understandable output for the application and then run the request.

#### Request Body:

```json
{
  "input": "<user input request>",
  "dryrun": true or false
}
```

* `input` (required): The user input requesting the AI to perform an action on some devices or get the devices list.
* `dryrun` (optional): If set to true, it will return the parameters to run the request but it won't run it. default is false.

#### Response:

- A JSON array containing the action to run and the parameters and the result of the action.

## Automation module
The automation module
### Standalone module
```
import connpy
router = connpy.node("uniqueName","ip/host", user="user", password="pass")
router.run(["term len 0","show run"])
print(router.output)
hasip = router.test("show ip int brief","1.1.1.1")
if hasip:
    print("Router has ip 1.1.1.1")
else:
    print("router does not have ip 1.1.1.1")
```

### Using manager configuration
```
import connpy
conf = connpy.configfile()
device = conf.getitem("server@office")
server = connpy.node("unique name", **device, config=conf)
result = server.run(["cd /", "ls -la"])
print(result)
```
### Running parallel tasks 
```
import connpy
conf = connpy.configfile()
#You can get the nodes from the config from a folder and fitlering in it
nodes = conf.getitem("@office", ["router1", "router2", "router3"])
#You can also get each node individually:
nodes = {}
nodes["router1"] = conf.getitem("router1@office")
nodes["router2"] = conf.getitem("router2@office")
nodes["router10"] = conf.getitem("router10@datacenter")
#Also, you can create the nodes manually:
nodes = {}
nodes["router1"] = {"host": "1.1.1.1", "user": "user", "password": "pass1"}
nodes["router2"] = {"host": "1.1.1.2", "user": "user", "password": "pass2"}
nodes["router3"] = {"host": "1.1.1.2", "user": "user", "password": "pass3"}
#Finally you run some tasks on the nodes
mynodes = connpy.nodes(nodes, config = conf)
result = mynodes.test(["show ip int br"], "1.1.1.2")
for i in result:
    print("---" + i + "---")
    print(result[i])
    print()
# Or for one specific node
mynodes.router1.run(["term len 0". "show run"], folder = "/home/user/logs")
```
### Using variables
```
import connpy
config = connpy.configfile()
nodes = config.getitem("@office", ["router1", "router2", "router3"])
commands = []
commands.append("config t")
commands.append("interface lo {id}")
commands.append("ip add {ip} {mask}")
commands.append("end")
variables = {}
variables["router1@office"] = {"ip": "10.57.57.1"}
variables["router2@office"] = {"ip": "10.57.57.2"}
variables["router3@office"] = {"ip": "10.57.57.3"}
variables["__global__"] = {"id": "57"}
variables["__global__"]["mask"] =  "255.255.255.255"
expected = "!"
routers = connpy.nodes(nodes, config = config)
routers.run(commands, variables)
routers.test("ping {ip}", expected, variables)
for key in routers.result:
    print(key, ' ---> ', ("pass" if routers.result[key] else "fail"))
```
### Using AI
```
import connpy
conf = connpy.configfile()
organization = 'openai-org'
api_key = "openai-key"
myia = ai(conf, organization, api_key)
input = "go to router 1 and get me the full configuration"
result = myia.ask(input, dryrun = False)
print(result)
```
'''
from .core import node,nodes
from .configfile import configfile
from .connapp import connapp
from .api import *
from .ai import ai
from .plugins import Plugins
from ._version import __version__
from pkg_resources import get_distribution

__all__ = ["node", "nodes", "configfile", "connapp", "ai", "Plugins"]
__author__ = "Federico Luzzi"
__pdoc__ = {
    'core': False,
    'completion': False,
    'api': False,
    'plugins': False
}



#!/usr/bin/env python3
#Imports
import os
import re
from Crypto.PublicKey import RSA
from Crypto.Cipher import PKCS1_OAEP
import ast
import argparse
import sys
import inquirer
from .core import node,nodes
from ._version import __version__
from .api import start_api,stop_api,debug_api
from .ai import ai
from .plugins import Plugins
import yaml
import shutil
class NoAliasDumper(yaml.SafeDumper):
    def ignore_aliases(self, data):
        return True
import ast
from rich import print as mdprint
from rich.markdown import Markdown
try:
    from pyfzf.pyfzf import FzfPrompt
except:
    FzfPrompt = None



#functions and classes

class connapp:
    ''' This class starts the connection manager app. It's normally used by connection manager but you can use it on a script to run the connection manager your way and use a different configfile and key.
        '''

    def __init__(self, config):
        ''' 
            
        ### Parameters:  

            - config (obj): Object generated with configfile class, it contains
                            the nodes configuration and the methods to manage
                            the config file.

        '''
        self.node = node
        self.nodes = nodes
        self.start_api = start_api
        self.stop_api = stop_api
        self.debug_api = debug_api
        self.ai = ai
        self.config = config
        self.nodes_list = self.config._getallnodes()
        self.folders = self.config._getallfolders()
        self.profiles = list(self.config.profiles.keys())
        self.case = self.config.config["case"]
        try:
            self.fzf = self.config.config["fzf"]
        except:
            self.fzf = False


    def start(self,argv = sys.argv[1:]):
        ''' 
            
        ### Parameters:  

            - argv (list): List of arguments to pass to the app.
                           Default: sys.argv[1:]

        ''' 
        #DEFAULTPARSER
        defaultparser = argparse.ArgumentParser(prog = "conn", description = "SSH and Telnet connection manager", formatter_class=argparse.RawTextHelpFormatter)
        subparsers = defaultparser.add_subparsers(title="Commands", dest="subcommand")
        #NODEPARSER
        nodeparser = subparsers.add_parser("node", formatter_class=argparse.RawTextHelpFormatter) 
        nodecrud = nodeparser.add_mutually_exclusive_group()
        nodeparser.add_argument("node", metavar="node|folder", nargs='?', default=None, action=self._store_type, help=self._help("node"))
        nodecrud.add_argument("-v","--version", dest="action", action="store_const", help="Show version", const="version", default="connect")
        nodecrud.add_argument("-a","--add", dest="action", action="store_const", help="Add new node[@subfolder][@folder] or [@subfolder]@folder", const="add", default="connect")
        nodecrud.add_argument("-r","--del", "--rm", dest="action", action="store_const", help="Delete node[@subfolder][@folder] or [@subfolder]@folder", const="del", default="connect")
        nodecrud.add_argument("-e","--mod", "--edit", dest="action", action="store_const", help="Modify node[@subfolder][@folder]", const="mod", default="connect")
        nodecrud.add_argument("-s","--show", dest="action", action="store_const", help="Show node[@subfolder][@folder]", const="show", default="connect")
        nodecrud.add_argument("-d","--debug", dest="debug", action="store_true", help="Display all conections steps")
        nodeparser.add_argument("-t","--sftp", dest="sftp", action="store_true", help="Connects using sftp instead of ssh")
        nodeparser.set_defaults(func=self._func_node)
        #PROFILEPARSER
        profileparser = subparsers.add_parser("profile", description="Manage profiles") 
        profileparser.add_argument("profile", nargs=1, action=self._store_type, type=self._type_profile, help="Name of profile to manage")
        profilecrud = profileparser.add_mutually_exclusive_group(required=True)
        profilecrud.add_argument("-a", "--add", dest="action", action="store_const", help="Add new profile", const="add")
        profilecrud.add_argument("-r", "--del", "--rm", dest="action", action="store_const", help="Delete profile", const="del")
        profilecrud.add_argument("-e", "--mod", "--edit", dest="action", action="store_const", help="Modify profile", const="mod")
        profilecrud.add_argument("-s", "--show", dest="action", action="store_const", help="Show profile", const="show")
        profileparser.set_defaults(func=self._func_profile)
        #MOVEPARSER
        moveparser = subparsers.add_parser("move", aliases=["mv"], description="Move node") 
        moveparser.add_argument("move", nargs=2, action=self._store_type, help="Move node[@subfolder][@folder] dest_node[@subfolder][@folder]", default="move", type=self._type_node)
        moveparser.set_defaults(func=self._func_others)
        #COPYPARSER
        copyparser = subparsers.add_parser("copy", aliases=["cp"], description="Copy node") 
        copyparser.add_argument("cp", nargs=2, action=self._store_type, help="Copy node[@subfolder][@folder] new_node[@subfolder][@folder]", default="cp", type=self._type_node)
        copyparser.set_defaults(func=self._func_others)
        #LISTPARSER
        lsparser = subparsers.add_parser("list", aliases=["ls"], description="List profiles, nodes or folders") 
        lsparser.add_argument("ls", action=self._store_type, choices=["profiles","nodes","folders"], help="List profiles, nodes or folders", default=False)
        lsparser.add_argument("--filter", nargs=1, help="Filter results")
        lsparser.add_argument("--format", nargs=1, help="Format of the output of nodes using {name}, {NAME}, {location}, {LOCATION}, {host} and {HOST}")
        lsparser.set_defaults(func=self._func_others)
        #BULKPARSER
        bulkparser = subparsers.add_parser("bulk", description="Add nodes in bulk") 
        bulkparser.add_argument("bulk", const="bulk", nargs=0, action=self._store_type, help="Add nodes in bulk")
        bulkparser.set_defaults(func=self._func_others)
        # EXPORTPARSER
        exportparser = subparsers.add_parser("export", description="Export connection folder to Yaml file") 
        exportparser.add_argument("export", nargs="+", action=self._store_type, help="Export /path/to/file.yml [@subfolder1][@folder1] [@subfolderN][@folderN]")
        exportparser.set_defaults(func=self._func_export)
        # IMPORTPARSER
        importparser = subparsers.add_parser("import", description="Import connection folder to config from Yaml file") 
        importparser.add_argument("file", nargs=1, action=self._store_type, help="Import /path/to/file.yml")
        importparser.set_defaults(func=self._func_import)
        # AIPARSER
        aiparser = subparsers.add_parser("ai", description="Make request to an AI") 
        aiparser.add_argument("ask", nargs='*', help="Ask connpy AI something")
        aiparser.add_argument("--model", nargs=1, help="Set the OPENAI model id")
        aiparser.add_argument("--org", nargs=1, help="Set the OPENAI organization id")
        aiparser.add_argument("--api_key", nargs=1, help="Set the OPENAI API key")
        aiparser.set_defaults(func=self._func_ai)
        #RUNPARSER
        runparser = subparsers.add_parser("run", description="Run scripts or commands on nodes", formatter_class=argparse.RawTextHelpFormatter) 
        runparser.add_argument("run", nargs='+', action=self._store_type, help=self._help("run"), default="run")
        runparser.add_argument("-g","--generate", dest="action", action="store_const", help="Generate yaml file template", const="generate", default="run")
        runparser.set_defaults(func=self._func_run)
        #APIPARSER
        apiparser = subparsers.add_parser("api", description="Start and stop connpy api") 
        apicrud = apiparser.add_mutually_exclusive_group(required=True)
        apicrud.add_argument("-s","--start", dest="start", nargs="?", action=self._store_type, help="Start conppy api", type=int, default=8048, metavar="PORT")
        apicrud.add_argument("-r","--restart", dest="restart", nargs=0, action=self._store_type, help="Restart conppy api")
        apicrud.add_argument("-x","--stop", dest="stop", nargs=0, action=self._store_type, help="Stop conppy api")
        apicrud.add_argument("-d", "--debug", dest="debug", nargs="?", action=self._store_type, help="Run connpy server on debug mode", type=int, default=8048, metavar="PORT")
        apiparser.set_defaults(func=self._func_api)
        #PLUGINSPARSER
        pluginparser = subparsers.add_parser("plugin", description="Manage plugins") 
        plugincrud = pluginparser.add_mutually_exclusive_group(required=True)
        plugincrud.add_argument("--add", metavar=("PLUGIN", "FILE"), nargs=2, help="Add new plugin")
        plugincrud.add_argument("--update", metavar=("PLUGIN", "FILE"), nargs=2, help="Update plugin")
        plugincrud.add_argument("--del", dest="delete", metavar="PLUGIN", nargs=1, help="Delete plugin")
        plugincrud.add_argument("--enable", metavar="PLUGIN", nargs=1, help="Enable plugin")
        plugincrud.add_argument("--disable", metavar="PLUGIN", nargs=1, help="Disable plugin")
        plugincrud.add_argument("--list", dest="list", action="store_true", help="Disable plugin")
        pluginparser.set_defaults(func=self._func_plugin)
        #CONFIGPARSER
        configparser = subparsers.add_parser("config", description="Manage app config") 
        configcrud = configparser.add_mutually_exclusive_group(required=True)
        configcrud.add_argument("--allow-uppercase", dest="case", nargs=1, action=self._store_type, help="Allow case sensitive names", choices=["true","false"])
        configcrud.add_argument("--fzf", dest="fzf", nargs=1, action=self._store_type, help="Use fzf for lists", choices=["true","false"])
        configcrud.add_argument("--keepalive", dest="idletime", nargs=1, action=self._store_type, help="Set keepalive time in seconds, 0 to disable", type=int, metavar="INT")
        configcrud.add_argument("--completion", dest="completion", nargs=1, choices=["bash","zsh"], action=self._store_type, help="Get terminal completion configuration for conn")
        configcrud.add_argument("--configfolder", dest="configfolder", nargs=1, action=self._store_type, help="Set the default location for config file", metavar="FOLDER")
        configcrud.add_argument("--openai-org", dest="organization", nargs=1, action=self._store_type, help="Set openai organization", metavar="ORGANIZATION")
        configcrud.add_argument("--openai-api-key", dest="api_key", nargs=1, action=self._store_type, help="Set openai api_key", metavar="API_KEY")
        configcrud.add_argument("--openai-model", dest="model", nargs=1, action=self._store_type, help="Set openai model", metavar="MODEL")
        configparser.set_defaults(func=self._func_others)
        #Add plugins
        self.plugins = Plugins()
        try:
            core_path = os.path.dirname(os.path.realpath(__file__)) + "/core_plugins"
            self.plugins._import_plugins_to_argparse(core_path, subparsers)
        except:
            pass
        try:
            file_path = self.config.defaultdir + "/plugins"
            self.plugins._import_plugins_to_argparse(file_path, subparsers)
        except:
            pass
        for preload in self.plugins.preloads.values():
            preload.Preload(self)
        #Generate helps
        nodeparser.usage = self._help("usage", subparsers)
        nodeparser.epilog = self._help("end", subparsers)
        nodeparser.help = self._help("node")
        #Manage sys arguments
        self.commands = list(subparsers.choices.keys())
        profilecmds = []
        for action in profileparser._actions:
            profilecmds.extend(action.option_strings)
        if len(argv) >= 2 and argv[1] == "profile" and argv[0] in profilecmds:
            argv[1] = argv[0]
            argv[0] = "profile"
        if len(argv) < 1 or argv[0] not in self.commands:
            argv.insert(0,"node")
        args = defaultparser.parse_args(argv)
        if args.subcommand in self.plugins.plugins:
            self.plugins.plugins[args.subcommand].Entrypoint(args, self.plugins.plugin_parsers[args.subcommand].parser, self)
        else:
            return args.func(args)

    class _store_type(argparse.Action):
        #Custom store type for cli app.
        def __call__(self, parser, args, values, option_string=None):
            setattr(args, "data", values)
            delattr(args,self.dest)
            setattr(args, "command", self.dest)

    def _func_node(self, args):
        #Function called when connecting or managing nodes.
        if not self.case and args.data != None:
            args.data = args.data.lower()
        actions = {"version": self._version, "connect": self._connect, "add": self._add, "del": self._del, "mod": self._mod, "show": self._show}
        return actions.get(args.action)(args)

    def _version(self, args):
        print(__version__)

    def _connect(self, args):
        if args.data == None:
            matches = self.nodes_list
            if len(matches) == 0:
                print("There are no nodes created")
                print("try: conn --help")
                exit(9)
        else:
            if args.data.startswith("@"):
                matches = list(filter(lambda k: args.data in k, self.nodes_list))
            else:
                matches = list(filter(lambda k: k.startswith(args.data), self.nodes_list))
        if len(matches) == 0:
            print("{} not found".format(args.data))
            exit(2)
        elif len(matches) > 1:
            matches[0] = self._choose(matches,"node", "connect")
        if matches[0] == None:
            exit(7)
        node = self.config.getitem(matches[0])
        node = self.node(matches[0],**node, config = self.config)
        if args.sftp:
            node.protocol = "sftp"
        if args.debug:
            node.interact(debug = True)
        else:
            node.interact()

    def _del(self, args):
        if args.data == None:
            print("Missing argument node")
            exit(3)
        elif args.data.startswith("@"):
            matches = list(filter(lambda k: k == args.data, self.folders))
        else:
            matches = self.config._getallnodes(args.data)
        if len(matches) == 0:
            print("{} not found".format(args.data))
            exit(2)
        print("Removing: {}".format(matches))
        question = [inquirer.Confirm("delete", message="Are you sure you want to continue?")]
        confirm = inquirer.prompt(question)
        if confirm == None:
            exit(7)
        if confirm["delete"]:
            if args.data.startswith("@"):
                uniques = self.config._explode_unique(matches[0])
                self.config._folder_del(**uniques)
            else:
                for node in matches:
                    nodeuniques = self.config._explode_unique(node)
                    self.config._connections_del(**nodeuniques)
            self.config._saveconfig(self.config.file)
            if len(matches) == 1:
                print("{} deleted succesfully".format(matches[0]))
            else:
                print(f"{len(matches)} nodes deleted succesfully")

    def _add(self, args):
        args.data = self._type_node(args.data)
        if args.data == None:
            print("Missing argument node")
            exit(3)
        elif args.data.startswith("@"):
            type = "folder"
            matches = list(filter(lambda k: k == args.data, self.folders))
            reversematches = list(filter(lambda k: "@" + k == args.data, self.nodes_list))
        else:
            type = "node"
            matches = list(filter(lambda k: k == args.data, self.nodes_list))
            reversematches = list(filter(lambda k: k == "@" + args.data, self.folders))
        if len(matches) > 0:
            print("{} already exist".format(matches[0]))
            exit(4)
        if len(reversematches) > 0:
            print("{} already exist".format(reversematches[0]))
            exit(4)
        else:
            if type == "folder":
                uniques = self.config._explode_unique(args.data)
                if uniques == False:
                    print("Invalid folder {}".format(args.data))
                    exit(5)
                if "subfolder" in uniques.keys():
                    parent = "@" + uniques["folder"]
                    if parent not in self.folders:
                        print("Folder {} not found".format(uniques["folder"]))
                        exit(2)
                self.config._folder_add(**uniques)
                self.config._saveconfig(self.config.file)
                print("{} added succesfully".format(args.data))
            if type == "node":
                nodefolder = args.data.partition("@")
                nodefolder = "@" + nodefolder[2]
                if nodefolder not in self.folders and nodefolder != "@":
                    print(nodefolder + " not found")
                    exit(2)
                uniques = self.config._explode_unique(args.data)
                if uniques == False:
                    print("Invalid node {}".format(args.data))
                    exit(5)
                print("You can use the configured setting in a profile using @profilename.")
                print("You can also leave empty any value except hostname/IP.")
                print("You can pass 1 or more passwords using comma separated @profiles")
                print("You can use this variables on logging file name: ${id} ${unique} ${host} ${port} ${user} ${protocol}")
                print("Some useful tags to set for automation are 'os', 'screen_length_command', and 'prompt'.")
                newnode = self._questions_nodes(args.data, uniques)
                if newnode == False:
                    exit(7)
                self.config._connections_add(**newnode)
                self.config._saveconfig(self.config.file)
                print("{} added succesfully".format(args.data))

    def _show(self, args):
        if args.data == None:
            print("Missing argument node")
            exit(3)
        matches = list(filter(lambda k: k == args.data, self.nodes_list))
        if len(matches) == 0:
            print("{} not found".format(args.data))
            exit(2)
        node = self.config.getitem(matches[0])
        for k, v in node.items():
            if isinstance(v, str):
                print(k + ": " + v)
            elif isinstance(v, list):
                print(k + ":")
                for i in v:
                    print("  - " + i)
            elif isinstance(v, dict):
                print(k + ":")
                for i,d in v.items():
                    print("  - " + i + ": " + d)

    def _mod(self, args):
        if args.data == None:
            print("Missing argument node")
            exit(3)
        matches = self.config._getallnodes(args.data)
        if len(matches) == 0:
            print("No connection found with filter: {}".format(args.data))
            exit(2)
        elif len(matches) == 1:
            uniques = self.config._explode_unique(args.data)
            unique = matches[0]
        else:
            uniques = {"id": None, "folder": None}
            unique = None
        print("Editing: {}".format(matches))
        node = {}
        for i in matches:
            node[i] = self.config.getitem(i)
        edits = self._questions_edit()
        if edits == None:
            exit(7)
        updatenode = self._questions_nodes(unique, uniques, edit=edits)
        if not updatenode:
            exit(7)
        if len(matches) == 1:
            uniques.update(node[matches[0]])
            uniques["type"] = "connection"
            if sorted(updatenode.items()) == sorted(uniques.items()):
                print("Nothing to do here")
                return
            else:
                self.config._connections_add(**updatenode)
                self.config._saveconfig(self.config.file)
                print("{} edited succesfully".format(args.data))
        else:
            for k in node:
                updatednode = self.config._explode_unique(k)
                updatednode["type"] = "connection"
                updatednode.update(node[k])
                editcount = 0
                for key, should_edit in edits.items():
                    if should_edit:
                        editcount += 1
                        updatednode[key] = updatenode[key]
                if not editcount:
                    print("Nothing to do here")
                    return
                else:
                    self.config._connections_add(**updatednode)
            self.config._saveconfig(self.config.file)
            print("{} edited succesfully".format(matches))
            return


    def _func_profile(self, args):
        #Function called when managing profiles
        if not self.case:
            args.data[0] = args.data[0].lower()
        actions = {"add": self._profile_add, "del": self._profile_del, "mod": self._profile_mod, "show": self._profile_show}
        return actions.get(args.action)(args)

    def _profile_del(self, args):
        matches = list(filter(lambda k: k == args.data[0], self.profiles))
        if len(matches) == 0:
            print("{} not found".format(args.data[0]))
            exit(2)
        if matches[0] == "default":
            print("Can't delete default profile")
            exit(6)
        usedprofile = self.config._profileused(matches[0])
        if len(usedprofile) > 0:
            print("Profile {} used in the following nodes:".format(matches[0]))
            print(", ".join(usedprofile))
            exit(8)
        question = [inquirer.Confirm("delete", message="Are you sure you want to delete {}?".format(matches[0]))]
        confirm = inquirer.prompt(question)
        if confirm["delete"]:
            self.config._profiles_del(id = matches[0])
            self.config._saveconfig(self.config.file)
            print("{} deleted succesfully".format(matches[0]))

    def _profile_show(self, args):
        matches = list(filter(lambda k: k == args.data[0], self.profiles))
        if len(matches) == 0:
            print("{} not found".format(args.data[0]))
            exit(2)
        profile = self.config.profiles[matches[0]]
        for k, v in profile.items():
            if isinstance(v, str):
                print(k + ": " + v)
            elif isinstance(v, list):
                print(k + ":")
                for i in v:
                    print("  - " + i)
            elif isinstance(v, dict):
                print(k + ":")
                for i,d in v.items():
                    print("  - " + i + ": " + d)

    def _profile_add(self, args):
        matches = list(filter(lambda k: k == args.data[0], self.profiles))
        if len(matches) > 0:
            print("Profile {} Already exist".format(matches[0]))
            exit(4)
        newprofile = self._questions_profiles(args.data[0])
        if newprofile == False:
            exit(7)
        self.config._profiles_add(**newprofile)
        self.config._saveconfig(self.config.file)
        print("{} added succesfully".format(args.data[0]))

    def _profile_mod(self, args):
        matches = list(filter(lambda k: k == args.data[0], self.profiles))
        if len(matches) == 0:
            print("{} not found".format(args.data[0]))
            exit(2)
        profile = self.config.profiles[matches[0]]
        oldprofile = {"id": matches[0]}
        oldprofile.update(profile)
        edits = self._questions_edit()
        if edits == None:
            exit(7)
        updateprofile = self._questions_profiles(matches[0], edit=edits)
        if not updateprofile:
            exit(7)
        if sorted(updateprofile.items()) == sorted(oldprofile.items()):
            print("Nothing to do here")
            return
        else:
            self.config._profiles_add(**updateprofile)
            self.config._saveconfig(self.config.file)
            print("{} edited succesfully".format(args.data[0]))
    
    def _func_others(self, args):
        #Function called when using other commands
        actions = {"ls": self._ls, "move": self._mvcp, "cp": self._mvcp, "bulk": self._bulk, "completion": self._completion, "case": self._case, "fzf": self._fzf, "idletime": self._idletime, "configfolder": self._configfolder, "organization": self._openai, "api_key": self._openai, "model": self._openai}
        return actions.get(args.command)(args)

    def _ls(self, args):
        items = getattr(self, args.data)
        if args.filter:
            items = [ item for item in items if re.search(args.filter[0], item)]
        if args.format and args.data == "nodes":
            newitems = []
            for i in items:
                formated = {}
                info = self.config.getitem(i)
                if "@" in i:
                    name_part, location_part = i.split("@", 1)
                    formated["location"] = "@" + location_part
                else:
                    name_part = i
                    formated["location"] = ""
                formated["name"] = name_part
                formated["host"] = info["host"]
                items_copy = list(formated.items())
                for key, value in items_copy:
                    upper_key = key.upper()
                    upper_value = value.upper()
                    formated[upper_key] = upper_value
                newitems.append(args.format[0].format(**formated))
            items = newitems
        print(*items, sep="\n")

    def _mvcp(self, args):
        if not self.case:
            args.data[0] = args.data[0].lower()
            args.data[1] = args.data[1].lower()
        source = list(filter(lambda k: k == args.data[0], self.nodes_list))
        dest = list(filter(lambda k: k == args.data[1], self.nodes_list))
        if len(source) != 1:
            print("{} not found".format(args.data[0]))
            exit(2)
        if len(dest) > 0:
            print("Node {} Already exist".format(args.data[1]))
            exit(4)
        nodefolder = args.data[1].partition("@")
        nodefolder = "@" + nodefolder[2]
        if nodefolder not in self.folders and nodefolder != "@":
            print("{} not found".format(nodefolder))
            exit(2)
        olduniques = self.config._explode_unique(args.data[0])
        newuniques = self.config._explode_unique(args.data[1])
        if newuniques == False:
            print("Invalid node {}".format(args.data[1]))
            exit(5)
        node = self.config.getitem(source[0])
        newnode = {**newuniques, **node}
        self.config._connections_add(**newnode)
        if args.command == "move":
           self.config._connections_del(**olduniques) 
        self.config._saveconfig(self.config.file)
        action = "moved" if args.command == "move" else "copied"
        print("{} {} succesfully to {}".format(args.data[0],action, args.data[1]))

    def _bulk(self, args):
        newnodes = self._questions_bulk()
        if newnodes == False:
            exit(7)
        if not self.case:
            newnodes["location"] = newnodes["location"].lower()
            newnodes["ids"] = newnodes["ids"].lower()
        ids = newnodes["ids"].split(",")
        hosts = newnodes["host"].split(",")
        count = 0
        for n in ids:
            unique = n + newnodes["location"]
            matches = list(filter(lambda k: k == unique, self.nodes_list))
            reversematches = list(filter(lambda k: k == "@" + unique, self.folders))
            if len(matches) > 0:
                print("Node {} already exist, ignoring it".format(unique))
                continue
            if len(reversematches) > 0:
                print("Folder with name {} already exist, ignoring it".format(unique))
                continue
            newnode = {"id": n}
            if newnodes["location"] != "":
                location = self.config._explode_unique(newnodes["location"])
                newnode.update(location)
            if len(hosts) > 1:
                index = ids.index(n)
                newnode["host"] = hosts[index]
            else:
                newnode["host"] = hosts[0]
            newnode["protocol"] = newnodes["protocol"]
            newnode["port"] = newnodes["port"]
            newnode["options"] = newnodes["options"]
            newnode["logs"] = newnodes["logs"]
            newnode["tags"] = newnodes["tags"]
            newnode["jumphost"] = newnodes["jumphost"]
            newnode["user"] = newnodes["user"]
            newnode["password"] = newnodes["password"]
            count +=1
            self.config._connections_add(**newnode)
            self.nodes_list = self.config._getallnodes()
        if count > 0:
            self.config._saveconfig(self.config.file)
            print("Succesfully added {} nodes".format(count))
        else:
            print("0 nodes added")

    def _completion(self, args):
        if args.data[0] == "bash":
            print(self._help("bashcompletion"))
        elif args.data[0] == "zsh":
            print(self._help("zshcompletion"))

    def _case(self, args):
        if args.data[0] == "true":
            args.data[0] = True
        elif args.data[0] == "false":
            args.data[0] = False
        self._change_settings(args.command, args.data[0])

    def _fzf(self, args):
        if args.data[0] == "true":
            args.data[0] = True
        elif args.data[0] == "false":
            args.data[0] = False
        self._change_settings(args.command, args.data[0])

    def _idletime(self, args):
        if args.data[0] < 0:
            args.data[0] = 0
        self._change_settings(args.command, args.data[0])

    def _configfolder(self, args):
        if not os.path.isdir(args.data[0]):
            raise argparse.ArgumentTypeError(f"readable_dir:{args.data[0]} is not a valid path")
        else:
            pathfile = self.config.defaultdir + "/.folder"
            folder = os.path.abspath(args.data[0]).rstrip('/')
            with open(pathfile, "w") as f:
                f.write(str(folder))
            print("Config saved")
        
    def _openai(self, args):
        if "openai" in self.config.config:
            openaikeys = self.config.config["openai"]
        else:
            openaikeys = {}
        openaikeys[args.command] = args.data[0]
        self._change_settings("openai", openaikeys)


    def _change_settings(self, name, value):
        self.config.config[name] = value
        self.config._saveconfig(self.config.file)
        print("Config saved")

    def _func_plugin(self, args):
        if args.add:
            if not os.path.exists(args.add[1]):
                print("File {} dosn't exists.".format(args.add[1]))
                exit(14)
            if args.add[0].isalpha() and args.add[0].islower() and len(args.add[0]) <= 15:
                disabled_dest_file = os.path.join(self.config.defaultdir + "/plugins", args.add[0] + ".py.bkp")
                if args.add[0] in self.commands or os.path.exists(disabled_dest_file):
                    print("Plugin name can't be the same as other commands.")
                    exit(15)
                else:
                    check_bad_script = self.plugins.verify_script(args.add[1])
                    if check_bad_script:
                        print(check_bad_script)
                        exit(16)
                    else:
                        try:
                            dest_file = os.path.join(self.config.defaultdir + "/plugins", args.add[0] + ".py")
                            shutil.copy2(args.add[1], dest_file)
                            print(f"Plugin {args.add[0]} added succesfully.")
                        except Exception as e:
                            print(f"Failed importing plugin file. {e}")
                            exit(17)
            else:
                print("Plugin name should be lowercase letters up to 15 characters.")
                exit(15)
        elif args.update:
            if not os.path.exists(args.update[1]):
                print("File {} dosn't exists.".format(args.update[1]))
                exit(14)
            plugin_file = os.path.join(self.config.defaultdir + "/plugins", args.update[0] + ".py")
            disabled_plugin_file = os.path.join(self.config.defaultdir + "/plugins", args.update[0] + ".py.bkp")
            plugin_exist = os.path.exists(plugin_file)
            disabled_plugin_exist = os.path.exists(disabled_plugin_file)
            if plugin_exist or disabled_plugin_exist:
                check_bad_script = self.plugins.verify_script(args.update[1])
                if check_bad_script:
                    print(check_bad_script)
                    exit(16)
                else:
                    try:
                        disabled_dest_file = os.path.join(self.config.defaultdir + "/plugins", args.update[0] + ".py.bkp")
                        dest_file = os.path.join(self.config.defaultdir + "/plugins", args.update[0] + ".py")
                        if disabled_plugin_exist:
                            shutil.copy2(args.update[1], disabled_dest_file)
                        else:
                            shutil.copy2(args.update[1], dest_file)
                        print(f"Plugin {args.update[0]} updated succesfully.")
                    except Exception as e:
                        print(f"Failed updating plugin file. {e}")
                        exit(17)

            else:
                print("Plugin {} dosn't exist.".format(args.update[0]))
                exit(14)
        elif args.delete:
            plugin_file = os.path.join(self.config.defaultdir + "/plugins", args.delete[0] + ".py")
            disabled_plugin_file = os.path.join(self.config.defaultdir + "/plugins", args.delete[0] + ".py.bkp")
            plugin_exist = os.path.exists(plugin_file)
            disabled_plugin_exist = os.path.exists(disabled_plugin_file)
            if not plugin_exist and not disabled_plugin_exist:
                print("Plugin {} dosn't exist.".format(args.delete[0]))
                exit(14)
            question = [inquirer.Confirm("delete", message="Are you sure you want to delete {} plugin?".format(args.delete[0]))]
            confirm = inquirer.prompt(question)
            if confirm == None:
                exit(7)
            if confirm["delete"]:
                try:
                    if plugin_exist:
                        os.remove(plugin_file)
                    elif disabled_plugin_exist:
                        os.remove(disabled_plugin_file)
                    print(f"plugin {args.delete[0]} deleted succesfully.")
                except Exception as e:
                    print(f"Failed deleting plugin file. {e}")
                    exit(17)
        elif args.disable:
            plugin_file = os.path.join(self.config.defaultdir + "/plugins", args.disable[0] + ".py")
            disabled_plugin_file = os.path.join(self.config.defaultdir + "/plugins", args.disable[0] + ".py.bkp")
            if not os.path.exists(plugin_file) or os.path.exists(disabled_plugin_file):
                print("Plugin {} dosn't exist or it's disabled.".format(args.disable[0]))
                exit(14)
            try:
                os.rename(plugin_file, disabled_plugin_file)
                print(f"plugin {args.disable[0]} disabled succesfully.")
            except Exception as e:
                print(f"Failed disabling plugin file. {e}")
                exit(17)
        elif args.enable:
            plugin_file = os.path.join(self.config.defaultdir + "/plugins", args.enable[0] + ".py")
            disabled_plugin_file = os.path.join(self.config.defaultdir + "/plugins", args.enable[0] + ".py.bkp")
            if os.path.exists(plugin_file) or not os.path.exists(disabled_plugin_file):
                print("Plugin {} dosn't exist or it's enabled.".format(args.enable[0]))
                exit(14)
            try:
                os.rename(disabled_plugin_file, plugin_file)
                print(f"plugin {args.enable[0]} enabled succesfully.")
            except Exception as e:
                print(f"Failed enabling plugin file. {e}")
                exit(17)
        elif args.list:
            enabled_files = []
            disabled_files = []
            plugins = {}
        
            # Iterate over all files in the specified folder
            for file in os.listdir(self.config.defaultdir + "/plugins"):
                # Check if the file is a Python file
                if file.endswith('.py'):
                    enabled_files.append(os.path.splitext(file)[0])
                # Check if the file is a Python backup file
                elif file.endswith('.py.bkp'):
                    disabled_files.append(os.path.splitext(os.path.splitext(file)[0])[0])
            if enabled_files:
                plugins["Enabled"] = enabled_files
            if disabled_files:
                plugins["Disabled"] = disabled_files
            if plugins:
                print(yaml.dump(plugins, sort_keys=False))
            else:
                print("There are no plugins added.")




    def _func_import(self, args):
        if not os.path.exists(args.data[0]):
            print("File {} dosn't exist".format(args.data[0]))
            exit(14)
        print("This could overwrite your current configuration!")
        question = [inquirer.Confirm("import", message="Are you sure you want to import {} file?".format(args.data[0]))]
        confirm = inquirer.prompt(question)
        if confirm == None:
            exit(7)
        if confirm["import"]:
            try:
                with open(args.data[0]) as file:
                    imported = yaml.load(file, Loader=yaml.FullLoader)
            except:
                print("failed reading file {}".format(args.data[0]))
                exit(10)
            for k,v in imported.items():
                uniques = self.config._explode_unique(k)
                if "folder" in uniques:
                    folder = f"@{uniques['folder']}"
                    matches = list(filter(lambda k: k == folder, self.folders))
                    if len(matches) == 0:
                        uniquefolder = self.config._explode_unique(folder)
                        self.config._folder_add(**uniquefolder)
                if "subfolder" in uniques:
                    subfolder = f"@{uniques['subfolder']}@{uniques['folder']}"
                    matches = list(filter(lambda k: k == subfolder, self.folders))
                    if len(matches) == 0:
                        uniquesubfolder = self.config._explode_unique(subfolder)
                        self.config._folder_add(**uniquesubfolder)
                uniques.update(v)
                self.config._connections_add(**uniques)
            self.config._saveconfig(self.config.file)
            print("File {} imported succesfully".format(args.data[0]))
        return

    def _func_export(self, args):
        if os.path.exists(args.data[0]):
            print("File {} already exists".format(args.data[0]))
            exit(14)
        if len(args.data[1:]) == 0:
            foldercons = self.config._getallnodesfull(extract = False)
        else:
            for folder in args.data[1:]:
                matches = list(filter(lambda k: k == folder, self.folders))
                if len(matches) == 0 and folder != "@":
                    print("{} folder not found".format(folder))
                    exit(2)
            foldercons = self.config._getallnodesfull(args.data[1:], extract = False)
        with open(args.data[0], "w") as file:
            yaml.dump(foldercons, file, Dumper=NoAliasDumper, default_flow_style=False)
            file.close()
        print("File {} generated succesfully".format(args.data[0]))
        exit()
        return

    def _func_run(self, args):
        if len(args.data) > 1:
            args.action = "noderun"
        actions = {"noderun": self._node_run, "generate": self._yaml_generate, "run": self._yaml_run}
        return actions.get(args.action)(args)

    def _func_ai(self, args):
        arguments = {}
        if args.model:
            arguments["model"] = args.model[0]
        if args.org:
            arguments["org"] = args.org[0]
        if args.api_key:
            arguments["api_key"] = args.api_key[0]
        self.myai = self.ai(self.config, **arguments)
        if args.ask:
            input = " ".join(args.ask)
            request = self.myai.ask(input, dryrun = True)
            if not request["app_related"]:
                mdprint(Markdown(request["response"]))
                print("\r")
            else:
                if request["action"] == "list_nodes":
                    if request["filter"]:
                        nodes = self.config._getallnodes(request["filter"])
                    else:
                        nodes = self.config._getallnodes()
                    list = "\n".join(nodes)
                    print(list)
                else:
                    yaml_data = yaml.dump(request["task"])
                    confirmation = f"I'm going to run the following task:\n```{yaml_data}```"
                    mdprint(Markdown(confirmation))
                    question = [inquirer.Confirm("task", message="Are you sure you want to continue?")]
                    print("\r")
                    confirm = inquirer.prompt(question)
                    if confirm == None:
                        exit(7)
                    if confirm["task"]:
                        script = {}
                        script["name"] = "RESULT"
                        script["output"] = "stdout"
                        script["nodes"] = request["nodes"]
                        script["action"] = request["action"]
                        if "expected" in request:
                            script["expected"] = request["expected"]
                        script.update(request["args"])
                        self._cli_run(script)
        else:
            history = None
            mdprint(Markdown("**Chatbot**: Hi! How can I help you today?\n\n---"))
            while True:
                questions = [
                        inquirer.Text('message', message="User", validate=self._ai_validation),
                    ]
                answers = inquirer.prompt(questions)
                if answers == None:
                    exit(7)
                response, history = self._process_input(answers["message"], history)
                mdprint(Markdown(f"""**Chatbot**:\n{response}\n\n---"""))
        return


    def _ai_validation(self, answers, current, regex = "^.+$"):
        #Validate ai user chat.
        if not re.match(regex, current):
            raise inquirer.errors.ValidationError("", reason="Can't send empty messages")
        return True

    def _process_input(self, input, history):
        response = self.myai.ask(input , chat_history = history, dryrun = True)
        if not response["app_related"]:
            try:
                if not history:
                    history = []
                history.extend(response["chat_history"])
            except:
                if not history:
                    history = None
            return response["response"], history
        else:
            history = None
            if response["action"] == "list_nodes":
                if response["filter"]:
                    nodes = self.config._getallnodes(response["filter"])
                else:
                    nodes = self.config._getallnodes()
                list = "\n".join(nodes)
                response = f"```{list}\n```"
            else:
                yaml_data = yaml.dump(response["task"])
                confirmresponse = f"I'm going to run the following task:\n```{yaml_data}```\nPlease confirm"
                while True:
                    mdprint(Markdown(f"""**Chatbot**:\n{confirmresponse}"""))
                    questions = [
                            inquirer.Text('message', message="User", validate=self._ai_validation),
                        ]
                    answers = inquirer.prompt(questions)
                    if answers == None:
                        exit(7)
                    confirmation = self.myai.confirm(answers["message"])
                    if isinstance(confirmation, bool):
                        if not confirmation:
                            response = "Request cancelled"
                        else:
                            nodes = self.nodes(self.config.getitems(response["nodes"]), config = self.config)
                            if response["action"] == "run":
                                output = nodes.run(**response["args"])
                                response = ""
                            elif response["action"] == "test":
                                result = nodes.test(**response["args"])
                                yaml_result = yaml.dump(result,default_flow_style=False, indent=4)
                                output = nodes.output
                                response = f"This is the result for your test:\n```\n{yaml_result}\n```"
                            for k,v in output.items():
                                response += f"\n***{k}***:\n```\n{v}\n```\n"
                        break
            return response, history

    def _func_api(self, args):
        if args.command == "stop" or args.command == "restart":
            args.data = self.stop_api()
        if args.command == "start" or args.command == "restart":
            if args.data:
                self.start_api(args.data)
            else:
                self.start_api()
        if args.command == "debug":
            if args.data:
                self.debug_api(args.data)
            else:
                self.debug_api()
        return

    def _node_run(self, args):
        command = " ".join(args.data[1:])
        script = {}
        script["name"] = "Output"
        script["action"] = "run"
        script["nodes"] = args.data[0]
        script["commands"] = [command]
        script["output"] = "stdout"
        self._cli_run(script)

    def _yaml_generate(self, args):
        if os.path.exists(args.data[0]):
            print("File {} already exists".format(args.data[0]))
            exit(14)
        else:
            with open(args.data[0], "w") as file:
                file.write(self._help("generate"))
                file.close()
            print("File {} generated succesfully".format(args.data[0]))
            exit()

    def _yaml_run(self, args):
        try:
            with open(args.data[0]) as file:
                scripts = yaml.load(file, Loader=yaml.FullLoader)
        except:
            print("failed reading file {}".format(args.data[0]))
            exit(10)
        for script in scripts["tasks"]:
            self._cli_run(script)


    def _cli_run(self, script):
        args = {}
        try:
            action = script["action"]
            nodelist = script["nodes"]
            args["commands"] = script["commands"]
            output = script["output"]
            if action == "test":
                args["expected"] = script["expected"]
        except KeyError as e:
            print("'{}' is mandatory".format(e.args[0]))
            exit(11)
        nodes = self.config._getallnodes(nodelist)
        if len(nodes) == 0:
            print("{} don't match any node".format(nodelist))
            exit(2)
        nodes = self.nodes(self.config.getitems(nodes), config = self.config)
        stdout = False
        if output is None:
            pass
        elif output == "stdout":
            stdout = True
        elif isinstance(output, str) and action == "run":
            args["folder"] = output
        if "variables" in script:
            args["vars"] = script["variables"]
        if "vars" in script:
            args["vars"] = script["vars"]
        try:
            options = script["options"]
            thisoptions = {k: v for k, v in options.items() if k in ["prompt", "parallel", "timeout"]}
            args.update(thisoptions)
        except:
            options = None
        try:
            size = str(os.get_terminal_size())
            p = re.search(r'.*columns=([0-9]+)', size)
            columns = int(p.group(1))
        except:
            columns = 80
        if action == "run":
            nodes.run(**args)
            print(script["name"].upper() + "-" * (columns - len(script["name"])))
            for i in nodes.status.keys():
                print("   " + i + " " + "-" * (columns - len(i) - 13) + (" PASS(0)" if nodes.status[i] == 0 else " FAIL({})".format(nodes.status[i])))
                if stdout:
                    for line in nodes.output[i].splitlines():
                        print("      " + line)
        elif action == "test":
            nodes.test(**args)
            print(script["name"].upper() + "-" * (columns - len(script["name"])))
            for i in nodes.status.keys():
                print("   " + i + " " + "-" * (columns - len(i) - 13) + (" PASS(0)" if nodes.status[i] == 0 else " FAIL({})".format(nodes.status[i])))
                if nodes.status[i] == 0:
                    max_length = max(len(s) for s in nodes.result[i].keys())
                    for k,v in nodes.result[i].items():
                        print("     TEST for '{}'".format(k) +  " "*(max_length - len(k) + 1) + "--> " + str(v).upper())
                if stdout:
                    if nodes.status[i] == 0:
                        print("     " + "-" * (max_length + 21))
                    for line in nodes.output[i].splitlines():
                        print("      " + line)
        else:
            print("Wrong action '{}'".format(action))
            exit(13)

    def _choose(self, list, name, action):
        #Generates an inquirer list to pick
        if FzfPrompt and self.fzf:
            fzf = FzfPrompt(executable_path="fzf-tmux")
            if not self.case:
                fzf = FzfPrompt(executable_path="fzf-tmux -i")
            answer = fzf.prompt(list, fzf_options="-d 25%")
            if len(answer) == 0:
                return
            else:
                return answer[0]
        else:
            questions = [inquirer.List(name, message="Pick {} to {}:".format(name,action), choices=list, carousel=True)]
            answer = inquirer.prompt(questions)
            if answer == None:
                return
            else:
                return answer[name]

    def _host_validation(self, answers, current, regex = "^.+$"):
        #Validate hostname in inquirer when managing nodes
        if not re.match(regex, current):
            raise inquirer.errors.ValidationError("", reason="Host cannot be empty")
        if current.startswith("@"):
            if current[1:] not in self.profiles:
                raise inquirer.errors.ValidationError("", reason="Profile {} don't exist".format(current))
        return True

    def _profile_protocol_validation(self, answers, current, regex = "(^ssh$|^telnet$|^$)"):
        #Validate protocol in inquirer when managing profiles
        if not re.match(regex, current):
            raise inquirer.errors.ValidationError("", reason="Pick between ssh, telnet or leave empty")
        return True

    def _protocol_validation(self, answers, current, regex = "(^ssh$|^telnet$|^$|^@.+$)"):
        #Validate protocol in inquirer when managing nodes
        if not re.match(regex, current):
            raise inquirer.errors.ValidationError("", reason="Pick between ssh, telnet, leave empty or @profile")
        if current.startswith("@"):
            if current[1:] not in self.profiles:
                raise inquirer.errors.ValidationError("", reason="Profile {} don't exist".format(current))
        return True

    def _profile_port_validation(self, answers, current, regex = "(^[0-9]*$)"):
        #Validate port in inquirer when managing profiles
        if not re.match(regex, current):
            raise inquirer.errors.ValidationError("", reason="Pick a port between 1-65535, @profile o leave empty")
        try:
            port = int(current)
        except:
            port = 0
        if current != "" and not 1 <= int(port) <= 65535:
            raise inquirer.errors.ValidationError("", reason="Pick a port between 1-65535 or leave empty")
        return True

    def _port_validation(self, answers, current, regex = "(^[0-9]*$|^@.+$)"):
        #Validate port in inquirer when managing nodes
        if not re.match(regex, current):
            raise inquirer.errors.ValidationError("", reason="Pick a port between 1-65535, @profile or leave empty")
        try:
            port = int(current)
        except:
            port = 0
        if current.startswith("@"):
            if current[1:] not in self.profiles:
                raise inquirer.errors.ValidationError("", reason="Profile {} don't exist".format(current))
        elif current != "" and not 1 <= int(port) <= 65535:
            raise inquirer.errors.ValidationError("", reason="Pick a port between 1-65535, @profile o leave empty")
        return True

    def _pass_validation(self, answers, current, regex = "(^@.+$)"):
        #Validate password in inquirer
        profiles = current.split(",")
        for i in profiles:
            if not re.match(regex, i) or i[1:] not in self.profiles:
                raise inquirer.errors.ValidationError("", reason="Profile {} don't exist".format(i))
        return True

    def _tags_validation(self, answers, current):
        #Validation for Tags in inquirer when managing nodes
        if current.startswith("@"):
            if current[1:] not in self.profiles:
                raise inquirer.errors.ValidationError("", reason="Profile {} don't exist".format(current))
        elif current != "":
            isdict = False
            try:
                isdict = ast.literal_eval(current)
            except:
                pass
            if not isinstance (isdict, dict):
                raise inquirer.errors.ValidationError("", reason="Tags should be a python dictionary.".format(current))
        return True

    def _profile_tags_validation(self, answers, current):
        #Validation for Tags in inquirer when managing profiles
        if current != "":
            isdict = False
            try:
                isdict = ast.literal_eval(current)
            except:
                pass
            if not isinstance (isdict, dict):
                raise inquirer.errors.ValidationError("", reason="Tags should be a python dictionary.".format(current))
        return True

    def _jumphost_validation(self, answers, current):
        #Validation for Jumphost in inquirer when managing nodes
        if current.startswith("@"):
            if current[1:] not in self.profiles:
                raise inquirer.errors.ValidationError("", reason="Profile {} don't exist".format(current))
        elif current != "":
            if current not in self.nodes_list :
                raise inquirer.errors.ValidationError("", reason="Node {} don't exist.".format(current))
        return True

    def _profile_jumphost_validation(self, answers, current):
        #Validation for Jumphost in inquirer when managing profiles
        if current != "":
            if current not in self.nodes_list :
                raise inquirer.errors.ValidationError("", reason="Node {} don't exist.".format(current))
        return True

    def _default_validation(self, answers, current):
        #Default validation type used in multiples questions in inquirer
        if current.startswith("@"):
            if current[1:] not in self.profiles:
                raise inquirer.errors.ValidationError("", reason="Profile {} don't exist".format(current))
        return True

    def _bulk_node_validation(self, answers, current, regex = "^[0-9a-zA-Z_.,$#-]+$"):
        #Validation of nodes when running bulk command
        if not re.match(regex, current):
            raise inquirer.errors.ValidationError("", reason="Host cannot be empty")
        if current.startswith("@"):
            if current[1:] not in self.profiles:
                raise inquirer.errors.ValidationError("", reason="Profile {} don't exist".format(current))
        return True

    def _bulk_folder_validation(self, answers, current):
        #Validation of folders when running bulk command
        if not self.case:
            current = current.lower()
        matches = list(filter(lambda k: k == current, self.folders))
        if current != "" and len(matches) == 0:
            raise inquirer.errors.ValidationError("", reason="Location {} don't exist".format(current))
        return True

    def _bulk_host_validation(self, answers, current, regex = "^.+$"):
        #Validate hostname when running bulk command
        if not re.match(regex, current):
            raise inquirer.errors.ValidationError("", reason="Host cannot be empty")
        if current.startswith("@"):
            if current[1:] not in self.profiles:
                raise inquirer.errors.ValidationError("", reason="Profile {} don't exist".format(current))
        hosts = current.split(",")
        nodes = answers["ids"].split(",")
        if len(hosts) > 1 and len(hosts) != len(nodes):
                raise inquirer.errors.ValidationError("", reason="Hosts list should be the same length of nodes list")
        return True

    def _questions_edit(self):
        #Inquirer questions when editing nodes or profiles
        questions = []
        questions.append(inquirer.Confirm("host", message="Edit Hostname/IP?"))
        questions.append(inquirer.Confirm("protocol", message="Edit Protocol?"))
        questions.append(inquirer.Confirm("port", message="Edit Port?"))
        questions.append(inquirer.Confirm("options", message="Edit Options?"))
        questions.append(inquirer.Confirm("logs", message="Edit logging path/file?"))
        questions.append(inquirer.Confirm("tags", message="Edit tags?"))
        questions.append(inquirer.Confirm("jumphost", message="Edit jumphost?"))
        questions.append(inquirer.Confirm("user", message="Edit User?"))
        questions.append(inquirer.Confirm("password", message="Edit password?"))
        answers = inquirer.prompt(questions)
        return answers

    def _questions_nodes(self, unique, uniques = None, edit = None):
        #Questions when adding or editing nodes
        try:
            defaults = self.config.getitem(unique)
            if "tags" not in defaults:
                defaults["tags"] = ""
            if "jumphost" not in defaults:
                defaults["jumphost"] = ""
        except:
            defaults = { "host":"", "protocol":"", "port":"", "user":"", "options":"", "logs":"" , "tags":"", "password":"", "jumphost":""}
        node = {}
        if edit == None:
            edit = { "host":True, "protocol":True, "port":True, "user":True, "password": True,"options":True, "logs":True, "tags":True, "jumphost":True }
        questions = []
        if edit["host"]:
            questions.append(inquirer.Text("host", message="Add Hostname or IP", validate=self._host_validation, default=defaults["host"]))
        else:
            node["host"] = defaults["host"]
        if edit["protocol"]:
            questions.append(inquirer.Text("protocol", message="Select Protocol", validate=self._protocol_validation, default=defaults["protocol"]))
        else:
            node["protocol"] = defaults["protocol"]
        if edit["port"]:
            questions.append(inquirer.Text("port", message="Select Port Number", validate=self._port_validation, default=defaults["port"]))
        else:
            node["port"] = defaults["port"]
        if edit["options"]:
            questions.append(inquirer.Text("options", message="Pass extra options to protocol", validate=self._default_validation, default=defaults["options"]))
        else:
            node["options"] = defaults["options"]
        if edit["logs"]:
            questions.append(inquirer.Text("logs", message="Pick logging path/file ",  validate=self._default_validation, default=defaults["logs"].replace("{","{{").replace("}","}}")))
        else:
            node["logs"] = defaults["logs"]
        if edit["tags"]:
            questions.append(inquirer.Text("tags", message="Add tags dictionary",  validate=self._tags_validation, default=str(defaults["tags"]).replace("{","{{").replace("}","}}")))
        else:
            node["tags"] = defaults["tags"]
        if edit["jumphost"]:
            questions.append(inquirer.Text("jumphost", message="Add Jumphost node",  validate=self._jumphost_validation, default=str(defaults["jumphost"]).replace("{","{{").replace("}","}}")))
        else:
            node["jumphost"] = defaults["jumphost"]
        if edit["user"]:
            questions.append(inquirer.Text("user", message="Pick username", validate=self._default_validation, default=defaults["user"]))
        else:
            node["user"] = defaults["user"]
        if edit["password"]:
            questions.append(inquirer.List("password", message="Password: Use a local password, no password or a list of profiles to reference?", choices=["Local Password", "Profiles", "No Password"]))
        else:
            node["password"] = defaults["password"]
        answer = inquirer.prompt(questions)
        if answer == None:
            return False
        if "password" in answer.keys():
            if answer["password"] == "Local Password":
                passq = [inquirer.Password("password", message="Set Password")]
                passa = inquirer.prompt(passq)
                if passa == None:
                    return False
                answer["password"] = self.encrypt(passa["password"])
            elif answer["password"] == "Profiles":
                passq = [(inquirer.Text("password", message="Set a @profile or a comma separated list of @profiles", validate=self._pass_validation))]
                passa = inquirer.prompt(passq)
                if passa == None:
                    return False
                answer["password"] = passa["password"].split(",")
            elif answer["password"] == "No Password":
                answer["password"] = ""
        if "tags" in answer.keys() and not answer["tags"].startswith("@") and answer["tags"]:
            answer["tags"] = ast.literal_eval(answer["tags"])
        result = {**uniques, **answer, **node}
        result["type"] = "connection"
        return result

    def _questions_profiles(self, unique, edit = None):
        #Questions when adding or editing profiles
        try:
            defaults = self.config.profiles[unique]
            if "tags" not in defaults:
                defaults["tags"] = ""
            if "jumphost" not in defaults:
                defaults["jumphost"] = ""
        except:
            defaults = { "host":"", "protocol":"", "port":"", "user":"", "options":"", "logs":"", "tags": "", "jumphost": ""}
        profile = {}
        if edit == None:
            edit = { "host":True, "protocol":True, "port":True, "user":True, "password": True,"options":True, "logs":True, "tags":True, "jumphost":True }
        questions = []
        if edit["host"]:
            questions.append(inquirer.Text("host", message="Add Hostname or IP", default=defaults["host"]))
        else:
            profile["host"] = defaults["host"]
        if edit["protocol"]:
            questions.append(inquirer.Text("protocol", message="Select Protocol", validate=self._profile_protocol_validation, default=defaults["protocol"]))
        else:
            profile["protocol"] = defaults["protocol"]
        if edit["port"]:
            questions.append(inquirer.Text("port", message="Select Port Number", validate=self._profile_port_validation, default=defaults["port"]))
        else:
            profile["port"] = defaults["port"]
        if edit["options"]:
            questions.append(inquirer.Text("options", message="Pass extra options to protocol", default=defaults["options"]))
        else:
            profile["options"] = defaults["options"]
        if edit["logs"]:
            questions.append(inquirer.Text("logs", message="Pick logging path/file ", default=defaults["logs"].replace("{","{{").replace("}","}}")))
        else:
            profile["logs"] = defaults["logs"]
        if edit["tags"]:
            questions.append(inquirer.Text("tags", message="Add tags dictionary",  validate=self._profile_tags_validation, default=str(defaults["tags"]).replace("{","{{").replace("}","}}")))
        else:
            profile["tags"] = defaults["tags"]
        if edit["jumphost"]:
            questions.append(inquirer.Text("jumphost", message="Add Jumphost node",  validate=self._profile_jumphost_validation, default=str(defaults["jumphost"]).replace("{","{{").replace("}","}}")))
        else:
            profile["jumphost"] = defaults["jumphost"]
        if edit["user"]:
            questions.append(inquirer.Text("user", message="Pick username", default=defaults["user"]))
        else:
            profile["user"] = defaults["user"]
        if edit["password"]:
            questions.append(inquirer.Password("password", message="Set Password"))
        else:
            profile["password"] = defaults["password"]
        answer = inquirer.prompt(questions)
        if answer == None:
            return False
        if "password" in answer.keys():
            if answer["password"] != "":
                answer["password"] = self.encrypt(answer["password"])
        if "tags" in answer.keys() and answer["tags"]:
            answer["tags"] = ast.literal_eval(answer["tags"])
        result = {**answer, **profile}
        result["id"] = unique
        return result

    def _questions_bulk(self):
        #Questions when using bulk command
        questions = []
        questions.append(inquirer.Text("ids", message="add a comma separated list of nodes to add", validate=self._bulk_node_validation))
        questions.append(inquirer.Text("location", message="Add a @folder, @subfolder@folder or leave empty", validate=self._bulk_folder_validation))
        questions.append(inquirer.Text("host", message="Add comma separated list of Hostnames or IPs", validate=self._bulk_host_validation))
        questions.append(inquirer.Text("protocol", message="Select Protocol", validate=self._protocol_validation))
        questions.append(inquirer.Text("port", message="Select Port Number", validate=self._port_validation))
        questions.append(inquirer.Text("options", message="Pass extra options to protocol", validate=self._default_validation))
        questions.append(inquirer.Text("logs", message="Pick logging path/file ", validate=self._default_validation))
        questions.append(inquirer.Text("tags", message="Add tags dictionary",  validate=self._tags_validation))
        questions.append(inquirer.Text("jumphost", message="Add Jumphost node",  validate=self._jumphost_validation))
        questions.append(inquirer.Text("user", message="Pick username", validate=self._default_validation))
        questions.append(inquirer.List("password", message="Password: Use a local password, no password or a list of profiles to reference?", choices=["Local Password", "Profiles", "No Password"]))
        answer = inquirer.prompt(questions)
        if answer == None:
            return False
        if "password" in answer.keys():
            if answer["password"] == "Local Password":
                passq = [inquirer.Password("password", message="Set Password")]
                passa = inquirer.prompt(passq)
                answer["password"] = self.encrypt(passa["password"])
            elif answer["password"] == "Profiles":
                passq = [(inquirer.Text("password", message="Set a @profile or a comma separated list of @profiles", validate=self._pass_validation))]
                passa = inquirer.prompt(passq)
                answer["password"] = passa["password"].split(",")
            elif answer["password"] == "No Password":
                answer["password"] = ""
        answer["type"] = "connection"
        if "tags" in answer.keys() and not answer["tags"].startswith("@") and answer["tags"]:
            answer["tags"] = ast.literal_eval(answer["tags"])
        return answer

    def _type_node(self, arg_value, pat=re.compile(r"^[0-9a-zA-Z_.$@#-]+$")):
        if arg_value == None:
            raise ValueError("Missing argument node")
        if not pat.match(arg_value):
            raise ValueError(f"Argument error: {arg_value}")
        return arg_value
    
    def _type_profile(self, arg_value, pat=re.compile(r"^[0-9a-zA-Z_.$#-]+$")):
        if not pat.match(arg_value):
            raise ValueError
        return arg_value

    def _help(self, type, parsers = None):
        #Store text for help and other commands
        if type == "node":
            return "node[@subfolder][@folder]\nConnect to specific node or show all matching nodes\n[@subfolder][@folder]\nShow all available connections globally or in specified path"
        if type == "usage":
            commands = []
            for subcommand, subparser in parsers.choices.items():
                if subparser.description != None:
                    commands.append(subcommand)
            commands = ",".join(commands)
            usage_help = f"conn [-h] [--add | --del | --mod | --show | --debug] [node|folder] [--sftp]\n       conn {{{commands}}} ..."
            return usage_help
        if type == "end":
            help_dict = {}
            for subcommand, subparser in parsers.choices.items():
                if subparser.description == None and help_dict:
                    previous_key = next(reversed(help_dict.keys()))
                    help_dict[f"{previous_key}({subcommand})"] = help_dict.pop(previous_key)
                else:
                    help_dict[subcommand] = subparser.description
                subparser.description = None
            commands_help = "Commands:\n"
            commands_help += "\n".join([f"  {cmd:<15} {help_text}" for cmd, help_text in help_dict.items() if help_text != None])
            return commands_help
        if type == "bashcompletion":
            return '''
#Here starts bash completion for conn
_conn()
{
        mapfile -t strings < <(connpy-completion-helper "bash" "${#COMP_WORDS[@]}" "${COMP_WORDS[@]}")
        local IFS=$'\t\n'
        local home_dir=$(eval echo ~)
        local last_word=${COMP_WORDS[-1]/\~/$home_dir}
        COMPREPLY=($(compgen -W "$(printf '%s' "${strings[@]}")" -- "$last_word"))
        if [ "$last_word" != "${COMP_WORDS[-1]}" ]; then
            COMPREPLY=(${COMPREPLY[@]/$home_dir/\~})
        fi
}

complete -o nospace -o nosort -F _conn conn
complete -o nospace -o nosort -F _conn connpy
#Here ends bash completion for conn
        '''
        if type == "zshcompletion":
            return '''
#Here starts zsh completion for conn
autoload -U compinit && compinit
_conn()
{
    local home_dir=$(eval echo ~)
    last_word=${words[-1]/\~/$home_dir}
    strings=($(connpy-completion-helper "zsh" ${#words} $words[1,-2] $last_word))
    for string in "${strings[@]}"; do
        #Replace the expanded home directory with ~
        if [ "$last_word" != "$words[-1]" ]; then
            string=${string/$home_dir/\~}
        fi
        if [[ "${string}" =~ .*/$ ]]; then
            # If the string ends with a '/', do not append a space
            compadd -Q -S '' -- "$string"
        else
            # If the string does not end with a '/', append a space
            compadd -Q -S ' ' -- "$string"
        fi
    done
}
compdef _conn conn
compdef _conn connpy
#Here ends zsh completion for conn
            '''
        if type == "run":
            return "node[@subfolder][@folder] commmand to run\nRun the specific command on the node and print output\n/path/to/file.yaml\nUse a yaml file to run an automation script"
        if type == "generate":
            return '''---
tasks:
- name: "Config"

  action: 'run' #Action can be test or run. Mandatory

  nodes: #List of nodes to work on. Mandatory
  - 'router1@office' #You can add specific nodes
  - '@aws'  #entire folders or subfolders
  - '@office':   #or filter inside a folder or subfolder
    - 'router2'
    - 'router7'

  commands: #List of commands to send, use {name} to pass variables
  - 'term len 0'
  - 'conf t'
  - 'interface {if}'
  - 'ip address 10.100.100.{id} 255.255.255.255'
  - '{commit}'
  - 'end'

  variables: #Variables to use on commands and expected. Optional
    __global__: #Global variables to use on all nodes, fallback if missing in the node.
      commit: ''
      if: 'loopback100'
    router1@office:
      id: 1
    router2@office:
      id: 2
      commit: 'commit'
    router3@office:
      id: 3
    vrouter1@aws:
      id: 4
    vrouterN@aws:
      id: 5
  
  output: /home/user/logs #Type of output, if null you only get Connection and test result. Choices are: null,stdout,/path/to/folder. Folder path only works on 'run' action.
  
  options:
    prompt: r'>$|#$|\$$|>.$|#.$|\$.$' #Optional prompt to check on your devices, default should work on most devices.
    parallel: 10 #Optional number of nodes to run commands on parallel. Default 10.
    timeout: 20 #Optional time to wait in seconds for prompt, expected or EOF. Default 20. 

- name: "TestConfig"
  action: 'test'
  nodes:
  - 'router1@office'
  - '@aws'
  - '@office':
    - 'router2'
    - 'router7'
  commands:
  - 'ping 10.100.100.{id}'
  expected: '!' #Expected text to find when running test action. Mandatory for 'test'
  variables:
    router1@office:
      id: 1
    router2@office:
      id: 2
      commit: 'commit'
    router3@office:
      id: 3
    vrouter1@aws:
      id: 4
    vrouterN@aws:
      id: 5
  output: null
...'''

    def encrypt(self, password, keyfile=None):
        '''
        Encrypts password using RSA keyfile

        ### Parameters:  

            - password (str): Plaintext password to encrypt.

        ### Optional Parameters:  

            - keyfile  (str): Path/file to keyfile. Default is config keyfile.
                              

        ### Returns:  

            str: Encrypted password.

        '''
        if keyfile is None:
            keyfile = self.config.key
        with open(keyfile) as f:
            key = RSA.import_key(f.read())
            f.close()
        publickey = key.publickey()
        encryptor = PKCS1_OAEP.new(publickey)
        password = encryptor.encrypt(password.encode("utf-8"))
        return str(password)




import sys
import os
import json
import glob
import importlib.util

def _getallnodes(config):
    #get all nodes on configfile
    nodes = []
    layer1 = [k for k,v in config["connections"].items() if isinstance(v, dict) and v["type"] == "connection"]
    folders = [k for k,v in config["connections"].items() if isinstance(v, dict) and v["type"] == "folder"]
    nodes.extend(layer1)
    for f in folders:
        layer2 = [k + "@" + f for k,v in config["connections"][f].items() if isinstance(v, dict) and v["type"] == "connection"]
        nodes.extend(layer2)
        subfolders = [k for k,v in config["connections"][f].items() if isinstance(v, dict) and v["type"] == "subfolder"]
        for s in subfolders:
            layer3 = [k + "@" + s + "@" + f for k,v in config["connections"][f][s].items() if isinstance(v, dict) and v["type"] == "connection"]
            nodes.extend(layer3)
    return nodes

def _getallfolders(config):
    #get all folders on configfile
    folders = ["@" + k for k,v in config["connections"].items() if isinstance(v, dict) and v["type"] == "folder"]
    subfolders = []
    for f in folders:
        s = ["@" + k + f for k,v in config["connections"][f[1:]].items() if isinstance(v, dict) and v["type"] == "subfolder"]
        subfolders.extend(s)
    folders.extend(subfolders)
    return folders

def _getcwd(words, option, folderonly=False):
    # Expand tilde to home directory if present
    if words[-1].startswith("~"):
        words[-1] = os.path.expanduser(words[-1])
    
    if words[-1] == option:
        path = './*'
    else:
        path = words[-1] + "*"

    pathstrings = glob.glob(path)
    for i in range(len(pathstrings)):
        if os.path.isdir(pathstrings[i]):
            pathstrings[i] += '/'
    pathstrings = [s[2:] if s.startswith('./') else s for s in pathstrings]
    if folderonly:
        pathstrings = [s for s in pathstrings if os.path.isdir(s)]
    return pathstrings

def _get_plugins(which, defaultdir):
    enabled_files = []
    disabled_files = []
    all_files = []
    all_plugins = {}
    # Iterate over all files in the specified folder
    for file in os.listdir(defaultdir + "/plugins"):
        # Check if the file is a Python file
        if file.endswith('.py'):
            enabled_files.append(os.path.splitext(file)[0])
            all_plugins[os.path.splitext(file)[0]] = os.path.join(defaultdir + "/plugins", file)
        # Check if the file is a Python backup file
        elif file.endswith('.py.bkp'):
            disabled_files.append(os.path.splitext(os.path.splitext(file)[0])[0])

    if which == "--disable":
        return enabled_files
    elif which == "--enable":
        return disabled_files
    elif which in ["--del", "--update"]:
        all_files.extend(enabled_files)
        all_files.extend(disabled_files)
        return all_files
    elif which == "all":
        return all_plugins


def main():
    home = os.path.expanduser("~")
    defaultdir = home + '/.config/conn'
    pathfile = defaultdir + '/.folder'
    try:
        with open(pathfile, "r") as f:
            configdir = f.read().strip()
    except:
        configdir = defaultdir
    defaultfile = configdir + '/config.json'
    jsonconf = open(defaultfile)
    config = json.load(jsonconf)
    nodes = _getallnodes(config)
    folders = _getallfolders(config)
    profiles = list(config["profiles"].keys())
    plugins = _get_plugins("all", defaultdir)
    info = {}
    info["config"] = config
    info["nodes"] = nodes
    info["folders"] = folders
    info["profiles"] = profiles
    info["plugins"] = plugins
    app = sys.argv[1]
    if app in ["bash", "zsh"]:
        positions = [2,4]
    else:
        positions = [1,3]
    wordsnumber = int(sys.argv[positions[0]])
    words = sys.argv[positions[1]:]
    if wordsnumber == 2:
        strings=["--add", "--del", "--rm", "--edit", "--mod", "--show", "mv", "move", "ls", "list", "cp", "copy", "profile", "run", "bulk", "config", "api", "ai", "export", "import", "--help", "plugin"]
        if plugins:
            strings.extend(plugins.keys())
        strings.extend(nodes)
        strings.extend(folders)

    elif wordsnumber >=3 and words[0] in plugins.keys():
        try:
            spec = importlib.util.spec_from_file_location("module.name", plugins[words[0]])
            module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(module)
            plugin_completion = getattr(module, "_connpy_completion")
            strings = plugin_completion(wordsnumber, words, info)
        except:
            exit()
    elif wordsnumber >= 3 and words[0] == "ai":
        if wordsnumber == 3:
            strings = ["--help", "--org", "--model", "--api_key"]
        else:
            strings = ["--org", "--model", "--api_key"]
    elif wordsnumber == 3:
        strings=[]
        if words[0] == "profile":
            strings=["--add", "--rm", "--del", "--edit", "--mod", "--show", "--help"]
        if words[0] == "config":
            strings=["--allow-uppercase", "--keepalive", "--completion", "--fzf", "--configfolder", "--openai-org", "--openai-org-api-key", "--openai-org-model","--help"]
        if words[0] == "api":
            strings=["--start", "--stop", "--restart", "--debug", "--help"]
        if words[0] in ["--mod", "--edit", "-e", "--show", "-s", "--add", "-a", "--rm", "--del", "-r"]:
            strings=["profile"]
        if words[0] in ["list", "ls"]:
            strings=["profiles", "nodes", "folders"]
        if words[0] in ["bulk", "mv", "cp", "copy"]:
            strings=["--help"]
        if words[0] in ["--rm", "--del", "-r"]:
            strings.extend(folders)
        if words[0] in ["--rm", "--del", "-r", "--mod", "--edit", "-e", "--show", "-s", "mv", "move", "cp", "copy"]:
            strings.extend(nodes)
        if words[0] == "plugin":
            strings = ["--help", "--add", "--update", "--del", "--enable", "--disable"]
        if words[0] in ["run", "import", "export"]:
            strings = ["--help"]
            if words[0] == "export":
                pathstrings = _getcwd(words, words[0], True)
            else:
                pathstrings = _getcwd(words, words[0])
            strings.extend(pathstrings)
            if words[0] == "run":
                strings.extend(nodes)

    elif wordsnumber >= 4 and words[0] == "export" and words[1] != "--help":
        strings = [item for item in folders if not any(word in item for word in words[:-1])]

    elif wordsnumber >= 4 and words[0] in ["list", "ls"] and words[1] == "nodes":
        options = ["--format", "--filter"]
        strings = [item for item in options if not any(word in item for word in words[:-1])]

    elif wordsnumber == 4:
          strings=[]
          if words[0] == "profile" and words[1] in ["--rm", "--del", "-r", "--mod", "--edit", "-e", "--show", "-s"]:
              strings.extend(profiles)
          if words[1] == "profile" and words[0] in ["--rm", "--del", "-r", "--mod", "--edit", "-e", "--show", "-s"]:
              strings.extend(profiles)
          if words[0] == "config" and words[1] == "--completion":
              strings=["bash", "zsh"]
          if words[0] == "config" and words[1] in ["--fzf", "--allow-uppercase"]:
              strings=["true", "false"]
          if words[0] == "config" and words[1] in ["--configfolder"]:
              strings=_getcwd(words,words[1],True)
          if words[0] == "plugin" and words[1] in ["--update", "--del", "--enable", "--disable"]:
              strings=_get_plugins(words[1], defaultdir)

    elif wordsnumber == 5 and words[0] == "plugin" and words[1] in ["--add", "--update"]:
            strings=_getcwd(words, words[2])
    else:
        exit()


    if app == "bash":
        strings = [s if s.endswith('/') else f"'{s} '" for s in strings]
    print('\t'.join(strings))

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python3
#Imports
from functools import wraps, partial

#functions and classes

class MethodHook:
    """Decorator class to enable Methods hooking"""

    def __init__(self, func):
        self.func = func
        self.pre_hooks = []   # List to store registered pre-hooks
        self.post_hooks = []  # List to store registered post-hooks
        wraps(func)(self)

    def __call__(self, *args, **kwargs):
        # Execute pre-hooks before the original function
        for hook in self.pre_hooks:
            try:
                args, kwargs = hook(*args, **kwargs)
            except Exception as e:
                print(f"{self.func.__name__} Pre-hook {hook.__name__} raised an exception: {e}")

        try:
            result = self.func(*args, **kwargs)

        finally:
            # Execute post-hooks after the original function
            for hook in self.post_hooks:
                try:
                    result = hook(*args, **kwargs, result=result)  # Pass result to hooks
                except Exception as e:
                    print(f"{self.func.__name__} Post-hook {hook.__name__} raised an exception: {e}")

        return result

    def __get__(self, instance, owner):
        if not instance:
            return self
        else:
            return self.make_callable(instance)

    def make_callable(self, instance):
        # Returns a callable that also allows access to hook registration methods
        callable_instance = partial(self.__call__, instance)
        callable_instance.register_pre_hook = self.register_pre_hook
        callable_instance.register_post_hook = self.register_post_hook
        return callable_instance

    def register_pre_hook(self, hook):
        """Register a function to be called before the original function"""
        self.pre_hooks.append(hook)

    def register_post_hook(self, hook):
        """Register a function to be called after the original function"""
        self.post_hooks.append(hook)

class ClassHook:
    """Decorator class to enable Class Modifying"""
    def __init__(self, cls):
        self.cls = cls
        # Initialize deferred class hooks if they don't already exist
        if not hasattr(cls, 'deferred_class_hooks'):
            cls.deferred_class_hooks = []

    def __call__(self, *args, **kwargs):
        instance = self.cls(*args, **kwargs)
        # Attach instance-specific modify method
        instance.modify = self.make_instance_modify(instance)
        # Apply any deferred modifications
        for hook in self.cls.deferred_class_hooks:
            hook(instance)
        return instance

    def __getattr__(self, name):
        """Delegate attribute access to the class."""
        return getattr(self.cls, name)

    def modify(self, modification_func):
        """Queue a modification to be applied to all future instances of the class."""
        self.cls.deferred_class_hooks.append(modification_func)

    def make_instance_modify(self, instance):
        """Create a modify function that is bound to a specific instance."""
        def modify_instance(modification_func):
            """Modify this specific instance."""
            modification_func(instance)
        return modify_instance

